from __future__ import annotations
import torch
import torch.nn.functional as F
import math
from ....ext import exllamav3_ext as ext
from ....util.progress import ProgressBar
from ....util.memory import free_mem, list_gpu_tensors
from ....util.hadamard import get_hadamard_dt
from ....util import cuda_sync_active
from ....util.tensor import save_tensor_image
from functools import lru_cache
import threading

# Constant
had_k, had_n = 128, 128
codebook_scale = 1.24371088

codebook_mcg_mult = 0xCBAC1FED
codebook_mul1_mult = 0x83DCD12D

@lru_cache
def tensor_core_perm(device):
    perm_a = [0] * 256
    for t in range(32):
        r0 = (t % 4) * 2
        r1 = r0 + 1
        r2 = r0 + 8
        r3 = r0 + 9
        c0 = t // 4
        c1 = c0 + 8
        perm_a[t * 8 + 0] = r0 * 16 + c0
        perm_a[t * 8 + 1] = r1 * 16 + c0
        perm_a[t * 8 + 2] = r2 * 16 + c0
        perm_a[t * 8 + 3] = r3 * 16 + c0
        perm_a[t * 8 + 4] = r0 * 16 + c1
        perm_a[t * 8 + 5] = r1 * 16 + c1
        perm_a[t * 8 + 6] = r2 * 16 + c1
        perm_a[t * 8 + 7] = r3 * 16 + c1
    return torch.tensor(perm_a, dtype = torch.int, device = device)


@lru_cache
def tensor_core_perm_i(device):
    return torch.argsort(tensor_core_perm(device))


@lru_cache
def get_temp_buffers(device, K: int):
    max_batch_size = 256
    temp_costs = torch.zeros((max_batch_size, 2, 65536 >> K), dtype = torch.half, device = device)
    temp_edges = torch.zeros((max_batch_size, 256, 65536 >> K), dtype = torch.short, device = device)
    return temp_costs, temp_edges


def quantize_tiles(tiles, quant_args: dict):
    tiles = tiles.contiguous()
    assert tiles.shape[1] == 256
    assert tiles.dtype == torch.float

    K = quant_args["K"]
    mcg = "mcg" in quant_args
    mul1 = "mul1" in quant_args
    quantized_tiles = torch.zeros_like(tiles)
    quantized_idx = torch.zeros_like(tiles, dtype = torch.short)
    temp_costs, temp_edges = get_temp_buffers(tiles.device, K)
    ext.quantize_tiles(
        tiles,
        quantized_tiles,
        quantized_idx,
        temp_costs,
        temp_edges,
        K,
        mcg,
        mul1,
    )
    return quantized_tiles, quantized_idx


@lru_cache
def get_quant_stream(device):
    return torch.cuda.Stream(device = device)


pinned_tiles: torch.Tensor | None = None
pinned_q_tiles: torch.Tensor | None = None
pinned_q_idx: torch.Tensor | None = None
def get_pinned(num_tiles: int):
    global pinned_tiles, pinned_q_tiles, pinned_q_idx
    if pinned_tiles is None or pinned_tiles.shape[0] < num_tiles:
        pinned_tiles = torch.empty((num_tiles, 256), device = "cpu", dtype = torch.float, pin_memory = True)
        pinned_q_tiles = torch.empty((num_tiles, 256), device = "cpu", dtype = torch.float, pin_memory = True)
        pinned_q_idx = torch.empty((num_tiles, 256), device = "cpu", dtype = torch.int16, pin_memory = True)
    return pinned_tiles[:num_tiles, :], pinned_q_tiles[:num_tiles, :], pinned_q_idx[:num_tiles, :]


def quantize_tiles_multigpu(tiles, quant_args: dict):
    devices = quant_args["devices"]
    if len(devices) == 1:
        return quantize_tiles(tiles, quant_args)

    # Get pinned buffers
    pin_tiles, pin_q_tiles, pin_q_idx = get_pinned(tiles.shape[0])

    # Copy input tiles to pinned memory. Input is always on the first device in the split
    copy_input_event = torch.cuda.Event(blocking = False)
    main_stream = get_quant_stream(devices[0])
    with torch.cuda.stream(main_stream):
        tiles = tiles.contiguous()
        pin_tiles.copy_(tiles, non_blocking = True)
        copy_input_event.record(main_stream)

    # Create split slices for input tiles, output tiles and output indices
    ratios = quant_args.get("device_ratios")
    if ratios:
        s = sum(ratios)
        split_sizes = [tiles.shape[0] * r / s for r in ratios]
        split_sizes = [round(s / 16) * 16 for s in split_sizes]
        split_sizes[-1] += tiles.shape[0] - sum(split_sizes)
    else:
        split_sizes = [tiles.shape[0] // len(devices)] * len(devices)
        split_sizes[-1] += tiles.shape[0] - sum(split_sizes)

    # Account for negative splits (edge case if too many GPUs and/or tensor too small)
    for i in range(len(split_sizes) - 2, -1, -1):
        if split_sizes[i + 1] < 0:
            split_sizes[i] += split_sizes[i + 1]
            split_sizes[i + 1] = 0

    pin_split_tiles = torch.split(pin_tiles, split_sizes)
    pin_split_q_tiles = torch.split(pin_q_tiles, split_sizes)
    pin_split_q_idx = torch.split(pin_q_idx, split_sizes)

    slice_done_events = []
    for i, device in enumerate(devices):

        stream = get_quant_stream(device)
        with torch.cuda.stream(stream):

            # Wait for input in host memory
            if i > 0:
                stream.wait_event(copy_input_event)

            if split_sizes[i] > 0:

                # Asynchronously copy the slice from the pinned buffer to device memory
                dev_tiles = pin_split_tiles[i].to(device, non_blocking = True)

                # Preallocate output tensors on the device.
                dev_q_tiles = torch.empty_like(dev_tiles, device = device)
                dev_q_idx = torch.empty_like(dev_tiles, dtype = torch.short, device = device)

                # Work buffers
                K = quant_args["K"]
                mcg = "mcg" in quant_args
                mul1 = "mul1" in quant_args
                temp_costs, temp_edges = get_temp_buffers(device, K)

                ext.quantize_tiles(
                    dev_tiles,
                    dev_q_tiles,
                    dev_q_idx,
                    temp_costs,
                    temp_edges,
                    K,
                    mcg,
                    mul1
                )

                # Async copy back to pinned memory
                pin_split_q_tiles[i].copy_(dev_q_tiles, non_blocking = True)
                pin_split_q_idx[i].copy_(dev_q_idx, non_blocking = True)

            # Finished slice
            evt = torch.cuda.Event(blocking = False)
            slice_done_events.append(evt)
            evt.record(stream)

    # Copy pinned buffers to original device
    with torch.cuda.stream(main_stream):
        for evt in slice_done_events:
            main_stream.wait_event(evt)
        q_tiles = torch.empty_like(tiles, device = devices[0])
        q_idx = torch.empty_like(tiles, dtype = torch.short, device = devices[0])
        q_tiles.copy_(pin_q_tiles, non_blocking = True)
        q_idx.copy_(pin_q_idx, non_blocking = True)

    return q_tiles, q_idx


def quantize_tiles_multigpu_sync(tiles, quant_args: dict):
    devices = quant_args["devices"]
    if len(devices) == 1:
        return quantize_tiles(tiles, quant_args)

    tiles = tiles.contiguous()

    split_sizes = [tiles.shape[0] // len(devices)] * len(devices)
    split_sizes[-1] += tiles.shape[0] - sum(split_sizes)
    split_tiles = torch.split(tiles, split_sizes)
    tiles_per_device = [chunk.to(device) for chunk, device in zip(split_tiles, devices)]
    torch.cuda.synchronize()

    q_tiles_per_device = []
    q_idx_per_device = []
    for dev_tiles, device in zip(tiles_per_device, devices):
        with torch.cuda.stream(get_quant_stream(device)):
            dev_q_tiles, dev_q_idx = quantize_tiles(dev_tiles, quant_args)
            q_tiles_per_device.append(dev_q_tiles)
            q_idx_per_device.append(dev_q_idx)

    for device in devices:
        torch.cuda.synchronize(device)

    q_tiles_per_device = [x.to(devices[0]) for x in q_tiles_per_device]
    q_idx_per_device = [x.to(devices[0]) for x in q_idx_per_device]
    quantized_tiles = torch.cat(q_tiles_per_device, dim = 0)
    quantized_idx = torch.cat(q_idx_per_device, dim = 0)
    return quantized_tiles, quantized_idx


def preapply_had_l(x: torch.Tensor, had_dim):
    k, n = x.shape
    x_dtype = x.dtype
    x = x.to(torch.float)
    had = get_hadamard_dt(had_dim, x.device, x.dtype, 1 / math.sqrt(had_dim))
    x = (had @ x.view(-1, had_dim, n)).view(k, n)
    x = x.to(x_dtype)
    return x


def preapply_had_r(x: torch.Tensor, had_dim):
    k, n = x.shape
    x_dtype = x.dtype
    x = x.to(torch.float)
    had = get_hadamard_dt(had_dim, x.device, x.dtype, 1 / math.sqrt(had_dim))
    x = (x.view(k, -1, had_dim) @ had).view(k, n)
    x = x.to(x_dtype)
    return x


def blockwise_preapply_had_l_(x: torch.Tensor, had_dim):
    k, n = x.shape
    assert k % had_dim == 0
    assert x.dtype == torch.float
    had = get_hadamard_dt(had_dim, x.device, x.dtype, 1 / math.sqrt(had_dim))
    num_blocks = k // had_dim
    for i in range(num_blocks):
        start = i * had_dim
        end = start + had_dim
        block = x[start:end, :]  # shape (had_dim, n)
        block_transformed = had @ block.view(had_dim, n)
        x[start:end, :] = block_transformed


def blockwise_preapply_had_r_(x: torch.Tensor, had_dim):
    k, n = x.shape
    assert n % had_dim == 0
    assert x.dtype == torch.float
    had = get_hadamard_dt(had_dim, x.device, x.dtype, 1 / math.sqrt(had_dim))
    num_blocks = n // had_dim
    for i in range(num_blocks):
        start = i * had_dim
        end = start + had_dim
        block = x[:, start:end]  # shape (k, had_dim)
        block_transformed = block @ had
        x[:, start:end] = block_transformed


def block_ldl(H: torch.Tensor, b: int, verbose: bool):

    n, _ = H.shape
    assert (n % b == 0)
    m = n // b

    # Cholesky factorization: H = L @ L.T
    # Try on GPU first
    try:
        retry_cpu = False
        L = torch.linalg.cholesky(H)
        # H is not needed after this, move to CPU. Then overwrite H's GPU storage with L, since we can't otherwise
        # free up that VRAM as the tensor is referenced by the parent frame
        H_cpu = H.cpu()
        H.copy_(L)  # VRAM copy, tiny overhead
        L = H
        H = H_cpu

    # Fall back on CPU factorization
    except Exception as e:
        if e.__class__.__name__ == "OutOfMemoryError" or "CUDA out of memory" in str(e) or "HIP out of memory" in str(e):
            retry_cpu = True
        else:
            raise e
    if retry_cpu:
        print(f" !! Out of memory on {str(H.device)}, trying CPU fallback")
        free_mem()
        H_cpu = H.cpu()
        L_cpu = torch.linalg.cholesky(H_cpu)
        # This is ugly, but overwrite H in VRAM to avoid allocating a new tensor, then replace reference with CPU copy
        H.copy_(L_cpu)
        del L_cpu
        L = H
        H = H_cpu

    # Get blocks along diagonal of L: DL.shape = (m, b, b)
    DL = torch.diagonal(L.reshape(m, b, m, b), dim1 = 0, dim2 = 2).permute(2, 0, 1)

    # Compute D as D[i] = DL[i] @ DL[i].T for each diagonal block i (don't actually end up needing this)
    # D = DL @ DL.transpose(1, 2)

    # Invert each diagonal block
    DL = torch.linalg.inv(DL)

    # Multiply each block's column with its inverse
    L = L.view(n, m, b)
    for i in range(m):
        L[:, i, :] = L[:, i, :] @ DL[i, :, :]  # TODO: Could maybe be L[m * b:, i, :]?
    L = L.reshape(n, n).contiguous()

    # Insert block identity matrices along the diagonal.
    # TODO: Figure out if this is necessary. Diagonal blocks should already be identities after previous step
    L_block = L.view(m, b, m, b).permute(0, 2, 1,3)
    dr = torch.arange(m)
    L_block[dr, dr] = torch.stack([torch.eye(b, device = L.device, dtype = H.dtype)] * m)

    return L, H  # , D.to(DL.device)


def ldlq(
    weight: torch.Tensor,
    L: torch.Tensor,
    quant_args: dict,
    pb: ProgressBar | None = None
):
    """
    :param weight:
        Input weights, shape (k, n). If device is "cpu", result is collected on CPU as well, saving a bunch of
        VRAM but adding a little PCIe overhead and many sync points

    :param L:
        LDL decomposition of regularized H

    :param quant_args:
        dict:
         - K: bitrate
         - buf_size_k: buffer size for LDLQ, along k

    :param pb:
        Optional ProgressPar to update, k // 16 steps

    :return:
        tuple:
         - quantized weight, shape (k, n)
         - indices (unpacked), shape (k // 16, n // 16, 256), uint16_t
    """

    devices = quant_args["devices"]
    for device in devices:
        torch.cuda.synchronize(device)
    main_stream = get_quant_stream(devices[0])
    with torch.cuda.stream(main_stream):

        devices = quant_args["devices"]
        device = L.device
        assert device == torch.device(devices[0])

        buffer_device = weight.device
        size_k, size_n = weight.shape  # Row-major
        assert size_k % 16 == 0
        assert size_n % 128 == 0
        tiles_k = size_k // 16
        tiles_n = size_n // 16

        buf_size_k = max(quant_args.get("buf_size_k", 128), 16)
        assert buf_size_k % 16 == 0
        assert size_n % buf_size_k == 0

        p_row = 0

        # Work buffers
        prod_cache = torch.zeros((size_k, size_n), dtype = torch.float, device = device)
        weight_q = torch.zeros((size_k, size_n), dtype = torch.float, device = buffer_device)
        encoded = torch.zeros((tiles_k, tiles_n, 256), dtype = torch.short, device = buffer_device)

        for j in range(size_k, 0, -buf_size_k):
            i = j - buf_size_k

            # Current span is rows i:j
            b_weight = weight[i:j].to(device)
            b_weight_q = weight_q[i:j] if device == buffer_device else \
                torch.zeros_like(weight_q[i:j], device = device)
            b_encoded = encoded[i // 16 : j // 16] if device == buffer_device else \
                torch.zeros_like(encoded[i // 16 : j // 16], device = device)
            b_prod_cache = prod_cache[i:j]
            b_L = L[i:j]

            # Iterate over rows of blocks in current span
            for bj in range(buf_size_k, 0, -16):
                bi = bj - 16

                # Error so far for the current span
                bb_err = b_weight[bj:] - b_weight_q[bj:]

                # Corresponding slice of LDL decomposition of H
                bb_L = b_L[bj:, i + bi:i + bj]

                # Input tiles for quantization
                compensation_term = b_prod_cache[bi:bj]
                compensation_term.addmm_(bb_L.T, bb_err,  alpha = 1.0, beta = 1.0)
                rows = b_weight[bi:bj] + compensation_term

                tiles = rows.reshape(16, tiles_n, 16).permute(1, 0, 2).reshape(tiles_n, 256)

                # Pre-permute to tensor core layout
                tiles = tiles[:, tensor_core_perm(device)]

                # Quantize
                quant_w, quant_i = quantize_tiles_multigpu(tiles, quant_args)

                # Undo permutation on reconstructed tiles, but keep indices in tensor core layout
                quant_w = quant_w[:, tensor_core_perm_i(device)]

                # Store result
                quant_w = quant_w.reshape(tiles_n, 16, 16).permute(1, 0, 2).reshape(16, size_n)
                b_weight_q[bi:bj] = quant_w
                b_encoded[bi // 16 : bj // 16] = quant_i.unsqueeze(0)

                # Update progress
                if pb:
                    p_row += 1
                    pb.update(p_row)

            # Collect output
            if device != buffer_device:
                weight_q[i:j] = b_weight_q.to(buffer_device)
                encoded[i // 16 : j // 16] = b_encoded.to(buffer_device)

            # Cache error term for the rest of the matrix
            b_err = b_weight - b_weight_q
            prod_cache.addmm_(b_L.T, b_err, alpha = 1.0, beta = 1.0)

        for device in devices:
            torch.cuda.synchronize(device)

    return weight_q, encoded


finalize_capture_H_mutex = threading.Lock()

def finalize_capture_H(H_data: dict, quant_args: dict, verbose: bool):
    with finalize_capture_H_mutex:

        # Unswap H
        if "H_swap_device" in H_data:
            H_data["H"] = H_data["H"].to(H_data["H_swap_device"])
            del H_data["H_swap_device"]

        H = H_data["H"]
        if H_data["finalized"]:
            return H, H_data["L"], H_data["su"], H_data["diag"]

        # Mean of samples summed up during forward pass
        H /= H_data["count"]

        # Regularize diagonal
        diag_mean = torch.diag(H).mean()
        idx = torch.arange(H.shape[0])
        H[idx, idx] += quant_args.get("sigma_reg", 0.025) * diag_mean

        # Some tests
        diag = H[idx, idx].clone()

        if verbose:
            print(f"     - H min/max: {H.min().item():.6f}   {H.max().item():.6f}")
            print(f"     - H mean/std: {H.mean().item():.6f}   {H.std().item():.6f}")
            print(f"     - H diag min/max: {diag.min():.6f}   {diag.max():.6f} ")

        # Random sign flips for input channel, fixed for the first linear layer to quantize with this H
        k = H.shape[0]
        su = (torch.randn(k, device = H.device).sign() + 1e-5).sign().to(torch.float).unsqueeze(1)
        H_data["su"] = su

        # Input had
        H *= su.T
        blockwise_preapply_had_r_(H, had_k)
        H *= su
        blockwise_preapply_had_l_(H, had_k)

        # Get block LDL decomposition of H, zero diagonal
        L, H = block_ldl(H, 16, verbose)
        dr = torch.arange(k)
        L[dr, dr] = 0
        H_data["L"] = L

        # H is no longer needed except to compute proxy error so move to CPU
        H = H.cpu()
        H_data["H"] = H.cpu()

        H_data["finalized"] = True
        H_data["diag"] = diag
        return H, L, su, diag


def pack_trellis(encoded: torch.Tensor, quant_args: dict) -> torch.Tensor:
    K = quant_args["K"]
    shape = encoded.shape
    assert len(shape) == 3 and shape[2] == 256
    assert encoded.dtype == torch.int16
    packed_shape = (shape[0], shape[1], 256 * K // 16)
    packed = torch.zeros(packed_shape, dtype = torch.int16, device = encoded.device)
    ext.pack_trellis(packed, encoded.contiguous(), K)
    # unpacked = torch.zeros_like(encoded)
    # ext.unpack_trellis(unpacked, packed, K)
    # assert torch.equal(unpacked, encoded)
    return packed


def pack_signs(signs: torch.Tensor, quant_args: dict) -> torch.Tensor:
    signs = signs.half().flatten().contiguous()
    assert signs.shape[0] % 16 == 0
    packed = torch.zeros(signs.shape[0] // 16, dtype = torch.int16, device = signs.device)
    ext.pack_signs(packed, signs)
    return packed


def g_scale_gss(
    weight_r: torch.Tensor,
    verbose: bool,
    quant_args: dict,
    width: int = 3,
    pb: ProgressBar = None
):
    # Select a sample of tiles along a wrapped diagonal (sampling from every row and column of tiles, hopefully
    # representative) and search for the global scale within given range that minimizes the direct quantization
    # error
    tiles = []
    tiles_k = weight_r.shape[0] // 16
    tiles_n = weight_r.shape[1] // 16
    for i in range(max(tiles_k, tiles_n)):
        for w in range(width):
            k = (i % tiles_k) * 16
            n = ((i + w) % tiles_n) * 16
            tile = weight_r[k : k + 16, n : n + 16].clone()
            tile = tile.view(256)
            tile = tile[tensor_core_perm(weight_r.device)]
            tiles.append(tile)
    tiles = torch.stack(tiles)

    devices = quant_args["devices"]
    for device in devices:
        torch.cuda.synchronize(device)

    main_stream = get_quant_stream(devices[0])
    # TODO: Figure out why Torch always initializes cuda:0 when exiting this CM, even when it's not used
    with torch.cuda.stream(main_stream):

        def test_scale(scale: float):
            quant_w, quant_i = quantize_tiles_multigpu(tiles * scale, quant_args)
            mse = ((quant_w / scale - tiles) ** 2).mean()
            return mse

        # Assume quantization error is a unimodal function of scale, golden section search to find minimum
        phi = (1 + math.sqrt(5)) / 2
        resphi = 2 - phi

        a, b = 0.1, 1.9
        tol = 0.01
        delta1 = abs(b - a)

        x1 = a + resphi * (b - a)
        x2 = b - resphi * (b - a)
        f1 = test_scale(x1)
        f2 = test_scale(x2)
        while abs(b - a) > tol:
            # if verbose:
                # print(f"     - gss: a = {a:.6f}, b = {b:.6f}")
            if f1 < f2:
                b = x2
                x2 = x1
                f2 = f1
                x1 = a + resphi * (b - a)
                f1 = test_scale(x1)
            else:
                a = x1
                x1 = x2
                f1 = f2
                x2 = b - resphi * (b - a)
                f2 = test_scale(x2)
            delta2 = abs(b - a)
            if pb:
                pb.update(100 - 100 * int(delta2 / delta1))

        best_scale = (a + b) / 2
        if verbose:
            print(f"     - gss: min = {best_scale:.6f}, mse: {(f1 + f2) / 2:.6f}")

    devices = quant_args["devices"]
    for device in devices:
        torch.cuda.synchronize(device)

    return best_scale, (f1 + f2) / 2


def block_rms(x: torch.Tensor, dim: int, keepdim: bool = False, blocksize: int = 32):
    # Compute blockwise x.square().mean(dim, keepdim).sqrt()
    n = x.size(dim)
    sq = None
    for block in torch.split(x, blocksize, dim = dim):
        block_sq = block.square().sum(dim = dim, keepdim = keepdim)
        if sq is None:
            sq = block_sq
        else:
            sq += block_sq
    mean_sq = sq / n
    return mean_sq.sqrt()


def block_rms_n(x: torch.Tensor, dim: int = 0, blocksize: int = 32):
    # Compute blockwise x.square().mean().sqrt()
    n = 0
    sq = None
    for block in torch.split(x, blocksize, dim = dim):
        block_sq = block.square().sum()
        n += block.numel()
        if sq is None:
            sq = block_sq
        else:
            sq += block_sq
    mean_sq = sq / n
    return mean_sq.sqrt()


def block_nmse(x: torch.Tensor, y: torch.Tensor, dim: int = 0, blocksize: int = 32):
    # Compute blockwise (x - y).square().mean().item() / y.square().mean().item()
    sq = None
    diff_sq = None
    for block_x, block_y in zip(torch.split(x, blocksize, dim = dim), torch.split(y, blocksize, dim = dim)):
        block_sq = block_y.square().sum()
        block_diff_sq = (block_x - block_y).square().sum()
        if sq is None:
            sq = block_sq
            diff_sq = block_diff_sq
        else:
            sq += block_sq
            diff_sq += block_diff_sq
    return diff_sq.item() / (sq.item() + 1e-20)


def regularize(
    weight: torch.Tensor,
    su: torch.Tensor,
    sv: torch.Tensor,
    quant_args: dict,
    verbose: bool,
    H_diag: torch.Tensor | None,
    pb: ProgressBar | None,
    skip_g_scale: bool = False
):
    force_out_scales = quant_args["apply_out_scales"]

    # dist_ref = torch.empty((512,), dtype = torch.float, device = weight.device)
    # dist_r = torch.empty_like(dist_ref)
    def jsd(h1, h2):
        m = (h1 + h2) / 2
        eps = 1e-12
        js = F.kl_div((h1 + eps).log(), m, reduction = "sum") + \
             F.kl_div((h2 + eps).log(), m, reduction = "sum")
        return js / 2

    # From experiments, it seems the deciding factor in when scaling output channels is beneficial is when
    # the input to the linear layer is very irregular. After some testing, set the cutoff at 15% of the RMS sum
    # on 2% of the channels
    # TODO: More science
    if H_diag is not None:
        diag = H_diag.sqrt()
        diag, _ = torch.sort(diag, descending = True)
        cutoff = diag.shape[0] // 50
        skew_factor = diag[:cutoff].sum() / diag.sum()
        if verbose:
            print(f"     - input state skew: {skew_factor.item():.6f}")

        if force_out_scales is None:
            apply_out_scales = skew_factor.item() < 0.15
        else:
            apply_out_scales = force_out_scales

    else:
        apply_out_scales = True if force_out_scales is None else force_out_scales

    # Apply output scales
    if apply_out_scales:
        out_channel_scales = block_rms(weight, dim = 0, keepdim = True)
        out_channel_scales /= out_channel_scales.mean()
        sv = (sv * out_channel_scales + 1e-10).float()
        if verbose:
            out_channel_std = out_channel_scales.std().item()
            out_channel_mean = out_channel_scales.mean().item()
            print(f"     - out ch scales std/mean: {out_channel_std:.6f}   {out_channel_mean:.6f}")

    # Output sign flips (and scales)
    weight /= sv

    # Output hadamard transform
    blockwise_preapply_had_r_(weight, had_n)

    # Input sign flips and scales
    in_channel_scales = block_rms(weight, dim = 1, keepdim = True)
    su = (su * in_channel_scales / (-codebook_scale) + 1e-10).float()  # mustn't be inplace
    weight /= su
    blockwise_preapply_had_l_(weight, had_k)

    # Determine best scale for matrix by test quantizing a sample of tiles along a wrapped diagonal
    if not skip_g_scale:
        g_scale, mse_scale = g_scale_gss(weight, False, quant_args, pb = pb)
    else:
        g_scale = 1.0
    weight *= g_scale
    su /= g_scale

    # ext.test_distribution(weight_os, dist_r, dist_ref, -3.8, 3.8)
    # js_os = jsd(dist_r, dist_ref)

    if verbose:
        print(f"     - su/sv std: {su.std().item():.6f}   {sv.std().item():.6f}")
        print(f"     - global scale: {g_scale:.6f}")
        print(f"     - sample mse: {mse_scale.item():.6f}")
        print(f"     - apply_out_scales: {str(apply_out_scales)}")

    return apply_out_scales, weight, g_scale, su, sv


def quantize_exl3(
    weight: torch.Tensor,
    H_data: dict,
    quant_args: dict,
    return_weight_q: bool,
    progress_str: str | None = None,
    verbose: bool = False,
    swap_to_device: torch.device | None = None,
    save_reg: str = None
):
    """
    :param weight:
        Input tensor, row major shape (in_features, out_features)

    :param H_data:
        Dictionary of hessian tensor and related data, as collected by Linear wrapper class. May be reused between
        linear layers with the same input (e.g. Q, K and V projections)

    :param quant_args:
        dict:
         - K: bitrate
         - seed: integer seed for random sign flips etc.
         - sigma_reg: regularization factor

    :param return_weight_q:
        Return quantized weight

    :param progress_str:
        Show progress bar during quantization

    :param verbose:
        Dump extra stats

    :param swap_to_device:
        If input tensor is on CPU, move to this device before quantization

    :param save_reg:
        Save regularized tensor as image to the provided path

    :return:
        tuple:
          - quantized weight
          - proxy error: trace(err @ H @ err.T) / (W @ H @ W.T)
          - quantized and packed tensors
    """

    progress_text = None if not progress_str else progress_str.replace("<step>", "Preparing")
    with (ProgressBar(progress_text, 100) as pb):

        assert weight.dtype == torch.float
        tiles_k = weight.shape[0] // 16

        if "seed" in quant_args:
            torch.manual_seed(quant_args["seed"])

        devices = quant_args["devices"]
        if weight.device != torch.device(devices[0]):
            weight = weight.to(devices[0])

        device = weight.device if swap_to_device is None else swap_to_device
        k, n = weight.shape

        # Get H, LDL decomp. and input/output sign flips
        H, L, su, H_diag = finalize_capture_H(H_data, quant_args, verbose)
        if H.is_cuda: H = H.to(device)
        if L.is_cuda: L = L.to(device)
        if su.is_cuda: su = su.to(device)
        if H_diag.is_cuda: H_diag = H_diag.to(device)
        sv = (torch.randn(n, device = device).sign() + 1e-5).sign().to(torch.float).unsqueeze(0)

        # Move stored L to CPU (if not already), move working L to device
        H_data["L"] = H_data["L"].cpu()
        L = L.to(device)

        if swap_to_device is not None:
            weight = weight.to(swap_to_device)
        if verbose:
            weight_copy = weight.cpu()
        weight_r = weight
        del weight

        if verbose:
            rms = block_rms_n(weight_r, dim = 0)
            print(f"     - input tensor rms: {rms:.6f}")

        # Regularization
        apply_out_scales, weight_r, g_scale, su, sv = regularize(
            weight_r,
            su,
            sv,
            quant_args,
            verbose,
            H_diag,
            pb
        )

        if save_reg:
            save_tensor_image(weight_r, save_reg)

        if verbose:
            rms = weight_r.square().mean().sqrt()
            print(f"     - regularized rms:  {rms:.6f}")

        progress_text = None if not progress_str else progress_str.replace("<step>", "Quantizing")
        pb.update(0)
        pb.new_task(progress_text, tiles_k)

        # Select device for work buffers (CPU is slower for small tensors but saves a lot of VRAM on big ones)
        # TODO: Use pynvml or mem_get_info to predict whether CPU buffer is needed
        if weight_r.numel() > 5e8:
            weight_r = weight_r.cpu()

        # Quantize
        weight_q, encoded_q = ldlq(weight_r, L, quant_args, pb)
        del L

        pb.update(tiles_k)

        # Metrics
        try:
            def block_trace(A, B, block_size = 1024):
                total = 0.0
                for j_start in range(0, B.shape[1], block_size):
                    j_end = min(j_start + block_size, B.shape[1])
                    B_block = B[:, j_start:j_end]
                    A_j_block = A[j_start:j_end, :]
                    partial = torch.einsum("ik,ij,jk->", A, B_block, A_j_block)
                    total += partial.item()
                return total
            E = weight_r - weight_q  # may run on CPU
            W = weight_r
            Hd = H.to(device)
            weight_r = None
            E = E.to(device)
            num = block_trace(E, Hd)
            E = None
            W = W.to(device)
            den = block_trace(W, Hd)
            W = None
            Hd = None
            proxy_err = num / max(den, 1e-8)
        except torch.OutOfMemoryError:
            weight_r = None
            E = None
            W = None
            Hd = None
            proxy_err = -1.0

        # free_mem()

        if return_weight_q or verbose:
            weight_q = weight_q.to(device)
            weight_q = preapply_had_l(weight_q, had_k)
            weight_q *= su
            weight_q = preapply_had_r(weight_q, had_n)
            weight_q *= sv

            if verbose:
                weight = weight_copy.to(device)
                nmse = block_nmse(weight_q, weight)
                print(f"     - quant nmse: {nmse:.6f}")

        # Compile packed tensor
        suh = su.flatten().contiguous().to(dtype = torch.half, copy = True)
        svh = sv.flatten().contiguous().to(dtype = torch.half, copy = True)
        trellis = pack_trellis(encoded_q.to(device), quant_args)

        out_tensors = {
            # "scale": weight_scale.to(dtype = torch.float, copy = True),
            # "su": pack_signs(su, quant_args),
            "suh": suh,
            # "sv": pack_signs(sv, quant_args),
            "svh": svh,
            "trellis": trellis,
        }

        # Safetensors doesn't know what to do with a torch.uint32 tensor. Anyway, since the multipliers are now
        # locked, the values in these tensors are never read, but they need to be present in the model files to
        # indicate which codebook to use during inference, per individual tensor.
        if quant_args.get("mcg"):
            out_tensors.update({
                "mcg": torch.tensor(codebook_mcg_mult, dtype = torch.uint32).view(torch.int)
            })
        if quant_args.get("mul1"):
            out_tensors.update({
                "mul1": torch.tensor(codebook_mul1_mult, dtype = torch.uint32).view(torch.int)
            })

        quant_args.update({
            "apply_out_scales": apply_out_scales,
            "g_scale": g_scale,
        })

    return weight_q, proxy_err, out_tensors


from __future__ import annotations
from typing import override
import os, json
import numpy as np
import torch
import torch.nn.functional as F
from ..model.config import Config
from ..model.model import Model
from ..util.rope import RopeSettings, RopeStyle
from ..util.file import read_dict, no_default
from ..util.vision import size_to_longest_edge_and_patch_size, convert_to_rgb, normalize_image
from ..util.tensor import to2
from ..modules import (
    Module,
    RMSNorm,
    Embedding,
    TransformerBlock,
    Attention,
    GatedMLP,
    Linear,
    Conv,
    MLP,
)
from ..modules.attn import prepare_for_attn
from ..tokenizer import Tokenizer, MMEmbedding
from types import SimpleNamespace
from PIL import Image
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .ministral3 import Ministral3Config

class Mistral3Config(Config):
    arch_string = "Mistral3ForConditionalGeneration"

    def __init__(
        self,
        directory: str,
        **kwargs,
    ):
        super().__init__(
            directory,
            {"text": Mistral3Model, "vision": Mistral3VisionModel},
            **kwargs
        )

        # Attention params
        self.head_dim = self.read_cfg(int, "text_config->head_dim", None)
        self.hidden_size = self.read_cfg(int, "text_config->hidden_size", no_default)
        self.num_q_heads = self.read_cfg(int, "text_config->num_attention_heads", no_default)
        self.num_kv_heads = self.read_cfg(int, "text_config->num_key_value_heads", self.num_q_heads)

        if not self.head_dim:
            self.head_dim = self.hidden_size // self.num_q_heads

        # MLP params
        self.assert_cfg(str, "text_config->hidden_act", "silu", True)
        self.intermediate_size = self.read_cfg(int, "text_config->intermediate_size", no_default)

        # Norms
        self.rms_norm_eps = self.read_cfg(float, "text_config->rms_norm_eps", no_default)

        # Layers
        self.num_hidden_layers = self.read_cfg(int, "text_config->num_hidden_layers", no_default)
        self.tie_word_embeddings = self.read_cfg(bool, "text_config->tie_word_embeddings", False)

        # RoPE
        text_cfg = self.read_cfg(dict, "text_config", no_default)
        self.rope_settings = self.read_rope_settings_default(RopeStyle.NEOX, config_dict = text_cfg)

        # Vision model settings
        def unpack_patch_size(patch_temp: dict | int):
            if isinstance(patch_temp, dict):
                h, w = (patch_temp.get(x) for x in ["height", "width"])
                assert w == h, f"Pixtral image preprocessor requires square patches, not {h} x {w}"
                patch_temp = h
            assert isinstance(patch_temp, int), "Unexpected type for patch_size"
            return patch_temp

        self.vision = SimpleNamespace()
        self.vision.head_dim = self.read_cfg(int, ["vision_config->head_dim"], no_default)
        self.vision.num_q_heads = self.read_cfg(int, ["vision_config->num_attention_heads"], no_default)
        self.vision.num_kv_heads = self.read_cfg(int, ["vision_config->num_key_value_heads"], self.vision.num_q_heads)
        self.vision.multimodal_projector_bias = self.read_cfg(bool, ["multimodal_projector_bias"], False)
        self.vision.hidden_size = self.read_cfg(int, ["vision_config->hidden_size"], no_default)
        self.vision.patch_size = unpack_patch_size(self.read_cfg(object, ["vision_config->patch_size"], int(14)))
        self.vision.num_hidden_layers = self.read_cfg(int, ["vision_config->num_hidden_layers"], 24)
        self.vision.intermediate_size = self.read_cfg(int, ["vision_config->intermediate_size"], no_default)
        self.vision.merger_intermediate_size = self.vision.intermediate_size
        self.vision.rms_norm_eps = self.rms_norm_eps
        self.vision.image_size = self.read_cfg(int, ["vision_config->image_size"], 1540)
        self.vision.spatial_merge_size = self.read_cfg(int, ["spatial_merge_size"], 1)
        self.vision.rope_theta = self.read_cfg(int, ["vision_config->rope_theta"], 10000.0)

        vision_cfg = self.read_cfg(dict, "vision_config", no_default)
        self.vision.rope_settings = self.read_rope_settings_default(RopeStyle.NEOX, config_dict = vision_cfg)

        self.vision.num_channels = 3
        self.vision.feature_layer = -1
        self.assert_cfg(int, "vision_config->num_channels", self.vision.num_channels, True)
        self.assert_cfg(int, "vision_feature_layer", self.vision.feature_layer, True)

        # Vision preprocessor
        prep_path = os.path.join(self.directory, "preprocessor_config.json")
        if os.path.exists(prep_path):
            with open(prep_path, encoding = "utf8") as f:
                read_prep_config = json.load(f)
        else:
            prep_path = os.path.join(self.directory, "processor_config.json")
            with open(prep_path, encoding = "utf8") as f:
                read_prep_config = json.load(f)
                read_prep_config = read_prep_config["image_processor"]
        image_processor_type = read_dict(read_prep_config, str, ["image_processor_type"], no_default)
        assert image_processor_type in ["PixtralImageProcessor", "PixtralImageProcessorFast"], \
            f"Wrong image processor type: {image_processor_type}"
        self.vision_pp = SimpleNamespace()
        self.vision_pp.image_mean = read_dict(read_prep_config, list, ["image_mean"], no_default)
        self.vision_pp.image_std = read_dict(read_prep_config, list, ["image_std"], no_default)
        self.vision_pp.resample = read_dict(read_prep_config, int, ["resample"], no_default)
        self.vision_pp.rescale_factor = read_dict(read_prep_config, float, ["rescale_factor"], no_default)
        self.vision_pp.size = read_dict(read_prep_config, dict, ["size"], no_default)
        self.vision_pp.patch_size = unpack_patch_size(read_dict(read_prep_config, object, ["patch_size"], no_default))

        assert self.vision.patch_size == self.vision_pp.patch_size, \
            "Vision model and vision preprocessor patch sizes do not match"


class Mistral3Model(Model):
    config_class = Mistral3Config

    def __init__(
        self,
        config: Mistral3Config | Ministral3Config,
        key_prefix = "language_model.",
        **kwargs
    ):
        super().__init__(config, **kwargs)

        self.modules += [
            Embedding(
                config = config,
                key = key_prefix + "model.embed_tokens",
                vocab_size = config.vocab_size,
                hidden_size = config.hidden_size,
            )
        ]

        self.first_block_idx = len(self.modules)

        self.modules += [
            TransformerBlock(
                config = config,
                key = key_prefix + f"model.layers.{idx}",
                attn_norm = RMSNorm(
                    config = config,
                    key = key_prefix + f"model.layers.{idx}.input_layernorm",
                    rms_norm_eps = config.rms_norm_eps,
                ),
                attn = Attention(
                    config = config,
                    key = key_prefix + f"model.layers.{idx}.self_attn",
                    layer_idx = idx,
                    hidden_size = config.hidden_size,
                    head_dim = config.head_dim,
                    num_q_heads = config.num_q_heads,
                    num_kv_heads = config.num_kv_heads,
                    rope_settings = config.rope_settings,
                    sm_scale = None,
                    key_q = "q_proj",
                    key_k = "k_proj",
                    key_v = "v_proj",
                    key_o = "o_proj",
                    qmap = "block.attn",
                    out_dtype = torch.float,
                ),
                mlp_norm = RMSNorm(
                    config = config,
                    key = key_prefix + f"model.layers.{idx}.post_attention_layernorm",
                    rms_norm_eps = config.rms_norm_eps,
                ),
                mlp = GatedMLP(
                    config = config,
                    key = key_prefix + f"model.layers.{idx}.mlp",
                    hidden_size = config.hidden_size,
                    intermediate_size = config.intermediate_size,
                    key_up = "up_proj",
                    key_gate = "gate_proj",
                    key_down = "down_proj",
                    qmap = "block.mlp",
                    # interm_dtype = torch.float,
                    out_dtype = torch.float,
                ),
            )
            for idx in range(config.num_hidden_layers)
        ]

        self.last_kv_module_idx = len(self.modules) - 1

        head_alt_key = None
        if config.tie_word_embeddings and not self.config.stc.has_tensor(key_prefix + "lm_head"):
            head_alt_key = key_prefix + "model.embed_tokens"

        self.modules += [
            RMSNorm(
                config = config,
                key = key_prefix + "model.norm",
                rms_norm_eps = config.rms_norm_eps,
                out_dtype = torch.half,
            ),
            Linear(
                config = config,
                key = key_prefix + "lm_head",
                qbits_key = "head_bits",
                alt_key = head_alt_key,
                in_features = config.hidden_size,
                out_features = config.vocab_size,
                qmap = "block",
                caps = {"logits_output": True}
            )
        ]

        self.logit_layer_idx = len(self.modules) - 1


    @override
    def prepare_inputs(self, input_ids: torch.Tensor, params: dict) -> torch.Tensor:
        input_ids = prepare_for_attn(input_ids, params)
        return input_ids


    @override
    def default_chat_prompt(self, prompt: str, system_prompt: str = None) -> str:
        p = "<s>"
        if system_prompt:
            p += f"[SYSTEM_PROMPT]{system_prompt}[/SYSTEM_PROMPT]"
        p += f"[INST]{prompt}[/INST]"
        return p


class Mistral3PatchMerger(Module):

    def __init__(
        self,
        config: Config,
        key: str,
        hidden_size: int,
        merge: int,
        out_dtype: torch.dtype | None = None,
        qmap: str | None = None,
    ):
        super().__init__(config, key, None)
        self.module_name = "Mistral3PatchMerger"
        self.qmap = qmap

        self.merge = merge
        self.hidden_size = hidden_size
        self.out_dtype = out_dtype

        self.merging_layer = Linear(
            config = config,
            key = f"{key}.merging_layer",
            in_features = hidden_size * merge ** 2,
            out_features = hidden_size
        )

        self.register_submodule(self.merging_layer)

    def optimizer_targets(self):
        raise NotImplementedError()

    @override
    def forward(
        self,
        x: torch.Tensor,
        params,
        out_dtype: torch.dtype | None = None,
    ) -> torch.Tensor:

        bsz, seq_len, dim = x.shape
        h, w = params["features_size"]
        assert bsz == 1

        x = x.view(h, w, dim).permute(2, 0, 1).unsqueeze(0)
        x = F.unfold(x, kernel_size = self.merge, stride = self.merge)
        x = x.view(bsz, dim * self.merge ** 2, -1).transpose(1, 2).contiguous()
        x = self.merging_layer.forward(x, params)

        return to2(x, out_dtype, self.out_dtype)


class Mistral3VisionModel(Model):

    @staticmethod
    @override
    def get_additional_compiled_tensors(config: Mistral3Config) -> dict:
        vlm_tensors = config.stc.list_tensors(prefix = "vision_tower")
        mmp_tensors = config.stc.list_tensors(prefix = "multi_modal_projector")
        return vlm_tensors | mmp_tensors

    def __init__(
        self,
        config: Mistral3Config,
        key_prefix = "vision_tower.",
        **kwargs
    ):
        super().__init__(config, **kwargs)
        self.config = config
        self.caps.update({"image_input": True})

        self.modules += [
            Conv(
                config = config,
                key = key_prefix + "patch_conv",
                in_channels = config.vision.num_channels,
                out_channels = config.vision.hidden_size,
                kernel_size = (config.vision.patch_size, config.vision.patch_size),
            ),
             RMSNorm(
                config = config,
                key = key_prefix + f"ln_pre",
                rms_norm_eps = config.rms_norm_eps,
            )
        ]

        self.modules += [
            TransformerBlock(
                config = config,
                key = key_prefix + f"transformer.layers.{idx}",
                attn_norm = RMSNorm(
                    config = config,
                    key = key_prefix + f"transformer.layers.{idx}.attention_norm",
                    rms_norm_eps = config.vision.rms_norm_eps
                ),
                attn = Attention(
                    config = config,
                    key = key_prefix + f"transformer.layers.{idx}.attention",
                    layer_idx = idx,
                    hidden_size = config.vision.hidden_size,
                    head_dim = config.vision.head_dim,
                    num_q_heads = config.vision.num_q_heads,
                    num_kv_heads = config.vision.num_kv_heads,
                    rope_settings = RopeSettings(
                        head_dim = config.vision.head_dim,
                        rope_theta = config.vision.rope_theta,
                    ),
                    key_q = "q_proj",
                    key_k = "k_proj",
                    key_v = "v_proj",
                    key_o = "o_proj",
                    qmap = "block.attn"
                ),
                mlp_norm = RMSNorm(
                    config = config,
                    key = key_prefix + f"transformer.layers.{idx}.ffn_norm",
                    rms_norm_eps = config.vision.rms_norm_eps
                ),
                mlp = GatedMLP(
                    config = config,
                    key = key_prefix + f"transformer.layers.{idx}.feed_forward",
                    hidden_size = config.vision.hidden_size,
                    intermediate_size = config.vision.intermediate_size,
                    key_gate = "gate_proj",
                    key_up = "up_proj",
                    key_down = "down_proj",
                    activation_fn = "silu",
                    qmap = "block.mlp",
                    pad_to = 1,
                ),
            )
            for idx in range(config.vision.num_hidden_layers)
        ]

        self.modules += [
            RMSNorm(
                config = config,
                key = f"multi_modal_projector.norm",
                rms_norm_eps = config.vision.rms_norm_eps,
                out_dtype = torch.half,
            ),
            Mistral3PatchMerger(
                config = config,
                key = "multi_modal_projector.patch_merger",
                hidden_size = config.vision.hidden_size,
                merge = config.vision.spatial_merge_size,
                out_dtype = torch.half,
            ),
            MLP(
                config = config,
                key = "multi_modal_projector",
                key_up = "linear_1",
                key_down = "linear_2",
                hidden_size = config.vision.hidden_size,
                intermediate_size = 5120,  # This seems to be hard coded?
                out_size = config.hidden_size,
                activation_fn = "gelu",
                qmap = "block",
            )
        ]

        # Precomputed RoPE table following Transformers implementation
        self.max_edge_features = config.vision_pp.size["longest_edge"] // config.vision_pp.patch_size
        freqs = 1.0 / (
            config.vision.rope_theta ** (
                torch.arange(0, config.vision.head_dim, 2).float()
                / config.vision.head_dim
            )
        )
        h = torch.arange(self.max_edge_features).float()
        w = torch.arange(self.max_edge_features).float()
        freqs_h = torch.outer(h, freqs[::2])
        freqs_w = torch.outer(w, freqs[1::2])
        self.inv_freq = torch.cat(
            [
                freqs_h[:, None, :].repeat(1, self.max_edge_features, 1),
                freqs_w[None, :, :].repeat(self.max_edge_features, 1, 1),
            ],
            dim = -1,
        ).reshape(-1, config.vision.head_dim // 2)


    def preprocess(
        self,
        image: Image
    ) -> (torch.Tensor, tuple):
        """
        Convert input image to the size and format expected by the vision tower. Image is scaled proportionally to
        fit a bounding box of longest_edge x longest_edge pixels as defined by the preprocessor config, while still
        being divisible into tiles of spatial_merge_size x spatial_merge_size input patches. Each such tile will be
        merged into one multimodal feature token by the vision tower.
        """

        patch_2d = (
            self.config.vision_pp.patch_size * self.config.vision.spatial_merge_size,
            self.config.vision_pp.patch_size * self.config.vision.spatial_merge_size,
        )
        longest_edge = self.config.vision_pp.size["longest_edge"]
        resample = Image.Resampling(self.config.vision_pp.resample)
        image_mean = tuple(self.config.vision_pp.image_mean)
        image_std = tuple(self.config.vision_pp.image_std)
        rescale_factor = self.config.vision_pp.rescale_factor

        # Convert to RGB and resize as necessary
        image = convert_to_rgb(image)
        old_size = image.size
        new_size = size_to_longest_edge_and_patch_size(image.size, (longest_edge, longest_edge), patch_2d)
        if old_size != new_size:
            image = image.resize(new_size, resample = resample)

        # Convert to numpy array and normalize
        image = np.array(image).astype(np.float32)
        image = image * rescale_factor
        image = normalize_image(image, image_mean, image_std)

        # Convert to tensor, shape (1, 3, resized_height, resized_width)
        image = image.transpose(2, 0, 1)
        image = torch.from_numpy(image).half().unsqueeze(0)
        return image, new_size


    def default_load_shape_dtype(self, chunk_size):
        return (
            (
                1,
                self.config.vision.num_channels,
                self.config.vision_pp.size["longest_edge"],
                self.config.vision_pp.size["longest_edge"]
            ),
            torch.half
        )


    def default_load_params(self):
        return {
            "features_size": (
                self.config.vision_pp.size["longest_edge"] // self.config.vision_pp.patch_size,
                self.config.vision_pp.size["longest_edge"] // self.config.vision_pp.patch_size,
            )
        }


    def get_image_embeddings(
        self,
        tokenizer: Tokenizer,
        image: Image | list[Image],
        text_alias: str | None = None,
    ):
        if isinstance(image, list):
            assert text_alias is None, "Cannot apply single alias to list of images"

            # Images in Mistral3 have uneven numbers of MM tokens so each image is actually processed at bsz 1
            return [self.get_image_embeddings(tokenizer, i) for i in image]

        image_tensor, prep_image_size = self.preprocess(image)
        features_size = (
            prep_image_size[1] // self.config.vision_pp.patch_size,
            prep_image_size[0] // self.config.vision_pp.patch_size,
        )

        # Flattened position ID grid matching inv_freq table
        h, w = features_size
        row_indices = torch.arange(h, dtype = torch.int).unsqueeze(1) * self.max_edge_features
        col_indices = torch.arange(w, dtype = torch.int).unsqueeze(0)
        position_ids_grid = (row_indices + col_indices).flatten().unsqueeze(0)

        embedding_tensor = self.forward(
            image_tensor,
            params = {
                "causal": False,
                "features_size": features_size,
                "inv_freq": self.inv_freq,
                "position_ids": position_ids_grid,
            }
        ).cpu().squeeze(0)

        w //= self.config.vision.spatial_merge_size
        h //= self.config.vision.spatial_merge_size
        id_break = tokenizer.single_id("[IMG_BREAK]")
        id_end = tokenizer.single_id("[IMG_END]")
        token_string = torch.tensor([([-1] * w + [id_break]) * h + [id_end]] , dtype = torch.long)

        mme = MMEmbedding(
            embeddings = embedding_tensor,
            text_alias = text_alias,
            token_string = token_string
        )

        mme.metadata.update({
            "original_size": image.size,
            "preprocessed_size": prep_image_size,
            "model_architecture": self.config.architecture,
        })

        return mme


    @override
    def prepare_inputs(self, input_ids: torch.Tensor, params: dict) -> torch.Tensor:
        return input_ids



from __future__ import annotations
from typing import override
import logging
import os
import re
import torch
import torch.nn.functional as F
from ..model.config import Config
from ..util.rope import RopeSettings, RoPE
from ..util.tensor import get_for_device, to2
from . import Module, Linear, RMSNorm, LayerNorm
from ..constants import PAGE_SIZE
from .multilinear import MultiLinear
from ..ext import exllamav3_ext as ext
from ..model.model_tp_alloc import TPAllocation
from ..util import profile_opt

try:
    from flash_attn import flash_attn_func, flash_attn_varlen_func, flash_attn_with_kvcache
except ImportError:
    flash_attn_func = None
    flash_attn_varlen_func = None
    flash_attn_with_kvcache = None

try:
    import flashinfer
except ImportError:
    flashinfer = None

_FLASHINFER_WORKSPACE_DEFAULT_BYTES = 128 * 1024 * 1024
_FLASHINFER_WORKSPACE_MIN_BYTES = 16 * 1024 * 1024
_FLASHINFER_WORKSPACE_ROUND_BYTES = 8 * 1024 * 1024
_FLASHINFER_WORKSPACE_GROWTH_MARGIN_BYTES = 16 * 1024 * 1024
_FLASHINFER_WORKSPACE: dict[tuple[str, int | None], torch.Tensor] = {}
_FLASHINFER_PREFILL_WRAPPERS: dict[
    tuple[str, int | None],
    flashinfer.BatchPrefillWithPagedKVCacheWrapper
] = {}
_FLASHINFER_DECODE_WRAPPERS: dict[
    tuple[str, int | None, str, bool],
    flashinfer.BatchDecodeWithPagedKVCacheWrapper
] = {}
_FLASHINFER_RUNTIME_STATS_ENABLED = os.getenv("EXLLAMAV3_RUNTIME_STATS", "").lower() in ("1", "true", "yes", "on")
_FLASHINFER_RUNTIME_STATS_INTERVAL = max(int(os.getenv("EXLLAMAV3_RUNTIME_STATS_INTERVAL", "256")), 1)
_FLASHINFER_RUNTIME_LOG = logging.getLogger("exllamav3.runtime.flashinfer")
_FLASHINFER_RUNTIME_STATS = {
    "decode_wrapper_creates": 0,
    "prefill_wrapper_creates": 0,
    "decode_plan_calls": 0,
    "decode_plan_reuses": 0,
    "prefill_plan_calls": 0,
    "prefill_plan_reuses": 0,
    "workspace_grows": 0,
}


def has_flash_attn_backend() -> bool:
    return (
        flash_attn_func is not None
        and flash_attn_varlen_func is not None
        and flash_attn_with_kvcache is not None
    )


def has_flashinfer_backend() -> bool:
    return flashinfer is not None


def _get_backend_policy_traits(config: object | None) -> dict[str, object]:
    config_dict = getattr(config, "config_dict", {}) or {}
    primary_cfg = config_dict
    for key in ("text_config", "language_config"):
        candidate = config_dict.get(key)
        if isinstance(candidate, dict):
            primary_cfg = candidate
            break

    def _get_int(key: str) -> int | None:
        value = primary_cfg.get(key, config_dict.get(key))
        return value if isinstance(value, int) and not isinstance(value, bool) else None

    architecture = getattr(config, "architecture", None)
    q_heads = _get_int("num_attention_heads")
    kv_heads = _get_int("num_key_value_heads")
    kv_lora_rank = _get_int("kv_lora_rank")
    sliding_window = _get_int("sliding_window")
    use_sliding_window = bool(
        primary_cfg.get("use_sliding_window", config_dict.get("use_sliding_window", False))
    )

    gqa_ratio = None
    if (
        q_heads is not None
        and kv_heads is not None
        and kv_heads > 0
        and q_heads % kv_heads == 0
    ):
        gqa_ratio = q_heads // kv_heads

    return {
        "architecture": architecture,
        "gqa_ratio": gqa_ratio,
        "mla_required": architecture == "DeepseekV2ForCausalLM" and kv_lora_rank is not None,
        "has_sliding_window": (
            (sliding_window is not None and sliding_window > 0) or use_sliding_window
        ),
    }


def resolve_auto_attention_backend(
    config: object | None,
    flash_attn_available: bool,
    flashinfer_available: bool,
) -> str:
    traits = _get_backend_policy_traits(config)
    architecture = traits["architecture"]
    gqa_ratio = traits["gqa_ratio"]
    mla_required = traits["mla_required"]
    has_sliding_window = traits["has_sliding_window"]

    # MLA support is a hard flashinfer requirement.
    if mla_required:
        if flashinfer_available:
            return "flashinfer"
        raise RuntimeError(
            "This model requires flashinfer backend for MLA support."
        )

    # Exaone4 with SWA is more reliable on SDPA than forcing flash-attn.
    if architecture == "Exaone4ForCausalLM" and has_sliding_window:
        return "sdpa"

    # Known flashinfer native non-tc decode gaps. Prefer flash-attn if present,
    # otherwise fall back to SDPA instead of forcing a slower/brittle path.
    if gqa_ratio in (5, 12):
        if flash_attn_available:
            return "flash_attn"
        return "sdpa"

    if flash_attn_available:
        return "flash_attn"
    if flashinfer_available:
        return "flashinfer"
    return "sdpa"


def _resolve_attn_mode(requested_mode: str, params: dict) -> str:
    if requested_mode in ("flash_attn", "flash_attn_nc") and not has_flash_attn_backend():
        raise RuntimeError("flash-attn backend requested but flash_attn is not installed")
    if requested_mode in ("flashinfer", "flashinfer_nc") and not has_flashinfer_backend():
        raise RuntimeError("flashinfer backend requested but flashinfer is not installed")

    if requested_mode == "auto":
        if has_flash_attn_backend():
            return "flash_attn"
        if has_flashinfer_backend():
            return "flashinfer"
        raise RuntimeError("No cache-capable attention backend is installed (need flash-attn or flashinfer)")

    if requested_mode == "auto_nc":
        if has_flash_attn_backend():
            return "flash_attn_nc"
        if has_flashinfer_backend():
            return "flashinfer_nc"
        return "sdpa_nc"

    return requested_mode


def _flashinfer_runtime_stat(key: str, amount: int = 1):
    if not _FLASHINFER_RUNTIME_STATS_ENABLED:
        return
    _FLASHINFER_RUNTIME_STATS[key] += amount
    total_plan_events = (
        _FLASHINFER_RUNTIME_STATS["decode_plan_calls"] +
        _FLASHINFER_RUNTIME_STATS["decode_plan_reuses"] +
        _FLASHINFER_RUNTIME_STATS["prefill_plan_calls"]
    )
    if total_plan_events and total_plan_events % _FLASHINFER_RUNTIME_STATS_INTERVAL == 0:
        decode_total = (
            _FLASHINFER_RUNTIME_STATS["decode_plan_calls"] +
            _FLASHINFER_RUNTIME_STATS["decode_plan_reuses"]
        )
        decode_reuse_rate = (
            _FLASHINFER_RUNTIME_STATS["decode_plan_reuses"] / decode_total
            if decode_total else 0.0
        )
        prefill_total = (
            _FLASHINFER_RUNTIME_STATS["prefill_plan_calls"] +
            _FLASHINFER_RUNTIME_STATS["prefill_plan_reuses"]
        )
        prefill_reuse_rate = (
            _FLASHINFER_RUNTIME_STATS["prefill_plan_reuses"] / prefill_total
            if prefill_total else 0.0
        )
        _FLASHINFER_RUNTIME_LOG.info(
            "flashinfer runtime stats: decode_plans=%d decode_reuses=%d "
            "decode_reuse_rate=%.2f prefill_plans=%d prefill_reuses=%d "
            "prefill_reuse_rate=%.2f decode_wrappers=%d "
            "prefill_wrappers=%d workspace_grows=%d",
            _FLASHINFER_RUNTIME_STATS["decode_plan_calls"],
            _FLASHINFER_RUNTIME_STATS["decode_plan_reuses"],
            decode_reuse_rate,
            _FLASHINFER_RUNTIME_STATS["prefill_plan_calls"],
            _FLASHINFER_RUNTIME_STATS["prefill_plan_reuses"],
            prefill_reuse_rate,
            _FLASHINFER_RUNTIME_STATS["decode_wrapper_creates"],
            _FLASHINFER_RUNTIME_STATS["prefill_wrapper_creates"],
            _FLASHINFER_RUNTIME_STATS["workspace_grows"],
        )


def _round_up_workspace_size(num_bytes: int) -> int:
    num_bytes = max(num_bytes, _FLASHINFER_WORKSPACE_MIN_BYTES)
    round_bytes = _FLASHINFER_WORKSPACE_ROUND_BYTES
    return ((num_bytes + round_bytes - 1) // round_bytes) * round_bytes


def _get_workspace_size_from_env() -> int:
    raw_bytes = os.getenv("EXLLAMAV3_FLASHINFER_WORKSPACE_BUFFER_SIZE")
    if raw_bytes is not None:
        try:
            return _round_up_workspace_size(int(raw_bytes))
        except ValueError:
            pass

    raw_mb = os.getenv("EXLLAMAV3_FLASHINFER_WORKSPACE_MB")
    if raw_mb is not None:
        try:
            return _round_up_workspace_size(int(raw_mb) * 1024 * 1024)
        except ValueError:
            pass

    return _round_up_workspace_size(_FLASHINFER_WORKSPACE_DEFAULT_BYTES)


_FLASHINFER_WORKSPACE_BYTES = _get_workspace_size_from_env()


def _parse_workspace_overflow_message(message: str) -> tuple[int, int] | None:
    match = re.search(
        r"with size\s+(\d+)\s+and alignment\s+\d+,\s+but only\s+(\d+)\s+bytes available",
        message,
    )
    if match is None:
        return None
    return int(match.group(1)), int(match.group(2))


def grow_flashinfer_workspace_for_exception(
    device: torch.device | None,
    exc: BaseException,
) -> bool:
    if device is None or not isinstance(exc, RuntimeError):
        return False
    parsed = _parse_workspace_overflow_message(str(exc))
    if parsed is None:
        return False

    required_bytes, available_bytes = parsed
    target_bytes = max(
        int(required_bytes * 1.25),
        required_bytes + _FLASHINFER_WORKSPACE_GROWTH_MARGIN_BYTES,
        available_bytes * 2,
    )
    target_bytes = _round_up_workspace_size(target_bytes)

    global _FLASHINFER_WORKSPACE_BYTES
    if target_bytes <= _FLASHINFER_WORKSPACE_BYTES:
        target_bytes = _round_up_workspace_size(_FLASHINFER_WORKSPACE_BYTES * 2)
    _FLASHINFER_WORKSPACE_BYTES = target_bytes

    key = (device.type, device.index)
    _FLASHINFER_WORKSPACE.pop(key, None)
    _FLASHINFER_PREFILL_WRAPPERS.pop(key, None)

    for wrapper_key in tuple(_FLASHINFER_DECODE_WRAPPERS.keys()):
        if wrapper_key[0] == device.type and wrapper_key[1] == device.index:
            _FLASHINFER_DECODE_WRAPPERS.pop(wrapper_key, None)

    _flashinfer_runtime_stat("workspace_grows")
    return True


def get_flashinfer_workspace(device: torch.device) -> torch.Tensor:
    if not has_flashinfer_backend():
        raise RuntimeError("flashinfer backend requested but flashinfer is not installed")
    key = (device.type, device.index)
    workspace = _FLASHINFER_WORKSPACE.get(key)
    if (
        workspace is None
        or workspace.device != device
        or workspace.numel() < _FLASHINFER_WORKSPACE_BYTES
    ):
        workspace = torch.empty(_FLASHINFER_WORKSPACE_BYTES, dtype = torch.uint8, device = device)
        _FLASHINFER_WORKSPACE[key] = workspace
    return workspace


def get_flashinfer_decode_wrapper_shared(
    device: torch.device,
    backend: str = "auto",
    use_tensor_cores: bool = False,
    plan_key: tuple | None = None,
) -> flashinfer.BatchDecodeWithPagedKVCacheWrapper:
    if not has_flashinfer_backend():
        raise RuntimeError("flashinfer backend requested but flashinfer is not installed")
    key = (device.type, device.index, backend, use_tensor_cores)
    if plan_key is not None:
        key = key + tuple(plan_key)
    wrapper = _FLASHINFER_DECODE_WRAPPERS.get(key)
    if wrapper is None:
        workspace = get_flashinfer_workspace(device)
        wrapper = flashinfer.BatchDecodeWithPagedKVCacheWrapper(
            workspace,
            kv_layout = "NHD",
            backend = backend,
            use_tensor_cores = use_tensor_cores,
        )
        _FLASHINFER_DECODE_WRAPPERS[key] = wrapper
        if _FLASHINFER_RUNTIME_STATS_ENABLED:
            _flashinfer_runtime_stat("decode_wrapper_creates")
    return wrapper


def get_flashinfer_prefill_wrapper_shared(
    device: torch.device,
) -> flashinfer.BatchPrefillWithPagedKVCacheWrapper:
    if not has_flashinfer_backend():
        raise RuntimeError("flashinfer backend requested but flashinfer is not installed")
    key = (device.type, device.index)
    wrapper = _FLASHINFER_PREFILL_WRAPPERS.get(key)
    if wrapper is None:
        workspace = get_flashinfer_workspace(device)
        wrapper = flashinfer.BatchPrefillWithPagedKVCacheWrapper(
            workspace,
            kv_layout = "NHD",
        )
        _FLASHINFER_PREFILL_WRAPPERS[key] = wrapper
        if _FLASHINFER_RUNTIME_STATS_ENABLED:
            _flashinfer_runtime_stat("prefill_wrapper_creates")
    return wrapper


def can_reuse_flashinfer_decode_plan(
    decode_wrapper: flashinfer.BatchDecodeWithPagedKVCacheWrapper,
    kv_indptr: torch.Tensor,
    kv_indices: torch.Tensor,
    kv_last_page_len: torch.Tensor,
) -> bool:
    if decode_wrapper.use_tensor_cores:
        return False

    paged_kv_indptr_buf = getattr(decode_wrapper, "_paged_kv_indptr_buf", None)
    paged_kv_indices_buf = getattr(decode_wrapper, "_paged_kv_indices_buf", None)
    paged_kv_last_page_len_buf = getattr(decode_wrapper, "_paged_kv_last_page_len_buf", None)
    plan_info = getattr(decode_wrapper, "_plan_info", None)
    if (
        plan_info is None
        or paged_kv_indptr_buf is None
        or paged_kv_indices_buf is None
        or paged_kv_last_page_len_buf is None
    ):
        return False
    if (
        paged_kv_indptr_buf.numel() != kv_indptr.numel()
        or paged_kv_indices_buf.numel() != kv_indices.numel()
        or paged_kv_last_page_len_buf.numel() != kv_last_page_len.numel()
    ):
        return False

    # flashinfer's non-tensor-core decode plan is tied to the per-request page
    # layout (`indptr`). The actual page indices can still change across
    # requests, so callers must refresh the wrapper's indices buffer even when
    # the plan itself can be reused.
    return torch.equal(paged_kv_indptr_buf, kv_indptr)


def make_paged_kv_metadata(
    block_table: torch.Tensor,
    kv_lens: torch.Tensor,
) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    kv_lens = kv_lens.to(dtype = torch.int32)
    num_pages = torch.div(kv_lens + PAGE_SIZE - 1, PAGE_SIZE, rounding_mode = "floor")
    block_table_i32 = block_table.to(dtype = torch.int32)
    if block_table_i32.shape[0] == 1:
        indices = block_table_i32[0, :int(num_pages[0].item())].contiguous()
    else:
        page_slices = [
            block_table_i32[row, :count]
            for row, count in enumerate(num_pages.tolist())
            if count > 0
        ]
        if page_slices:
            indices = torch.cat(page_slices, dim = 0).contiguous()
        else:
            indices = torch.empty((0,), dtype = torch.int32, device = block_table_i32.device)
    indptr = torch.empty((kv_lens.numel() + 1,), dtype = torch.int32, device = block_table.device)
    indptr[0] = 0
    indptr[1:] = torch.cumsum(num_pages, dim = 0)
    last_page_len = torch.where(
        kv_lens > 0,
        torch.remainder(kv_lens - 1, PAGE_SIZE) + 1,
        torch.ones_like(kv_lens)
    )
    return indptr.contiguous(), indices.contiguous(), last_page_len.contiguous()

"""
SDPA:
    
    attn_mode: "sdpa_nc"
    position (optional, default = 0): int *OR*
    positions: shape (bsz) *OR*
    position_ids: shape (bsz, seq_len)    
    - no cache
    - no chunking
    - batch shape is determined by shape of input_ids
    - no logit softcap support (Gemma)
                    
Flash Attention / FlashInfer:
                
    attn_mode: "flash_attn"
    attn_mode: "flashinfer"
    attn_mode: "auto"
    batch_shape: tuple of (bsz, max_seq_len)
    cache: Cache with capacity of at least bsz*max_seq_len tokens
    past_len: int, *OR*
    cache_seqlens: shape (bsz) 
    position: int (overrides past_len for position emb)
    positions: shape (bsz) (overrides cache_seqlens for position emb) *OR*
    position_ids: shape (bsz, seq_len) (overrides cache_seqlens for position emb)
    - max_seq_len must be divisible by 256
    
    attn_mode: "flash_attn"
    block_table: list of page indices, shape (bsz, pages_per_seq)
    cache: Paged cache
    cache_seqlens: shape (bsz)
    positions: shape (bsz) (overrides cache_seqlens for position emb) *OR*
    position_ids: shape (bsz, seq_len) (overrides cache_seqlens for position emb)

    attn_mode: "flash_attn_nc"
    attn_mode: "flashinfer_nc"
    attn_mode: "auto_nc"
    position (optional, default = 0): int *OR*
    positions: shape (bsz) *OR*
    position_ids: shape (bsz, seq_len)    
    - no cache
    - no chunking
    - batch shape is determined by shape of input_ids
"""

def prepare_sdpa_nc(input_ids: torch.Tensor, params: dict) -> torch.Tensor:
    assert "cache" not in params, \
        f"Cache provided for attn_mode: sdpa_nc"
    return input_ids


def prepare_flash_attn_nc(input_ids: torch.Tensor, params: dict) -> torch.Tensor:
    assert "cache" not in params, \
        f"Cache provided for attn_mode: sdpa_nc"
    return input_ids


def prepare_flashinfer_nc(input_ids: torch.Tensor, params: dict) -> torch.Tensor:
    return prepare_flash_attn_nc(input_ids, params)


def prepare_flash_attn(input_ids: torch.Tensor, params: dict) -> torch.Tensor:
    bsz, seq_len = input_ids.shape

    if "batch_shape" in params:
        cache = params["cache"]
        cache_bsz, cache_max_seq_len = params["batch_shape"]
        past_len = params.get("past_len")
        cache_seqlens = params.get("cache_seqlens") if past_len is None else None
        position = params.get("position") if past_len is None else None
        positions = params.get("positions") if past_len is None else None
        position_ids = params.get("position_ids") if past_len is None else None
        assert cache_bsz >= bsz, "batch size too large for cache"
        assert cache_max_seq_len % PAGE_SIZE == 0, f"cache seq len must be a multiple of {PAGE_SIZE}"
        # assert (past_len is not None) ^ (cache_seqlens is not None), "Need either past_len or cache_seqlens"
        assert bsz * cache_max_seq_len <= cache.max_num_tokens, "Cache too small for batch shape"
        cache_bsz = min(bsz, cache_bsz)
        num_pages = cache_bsz * cache_max_seq_len // PAGE_SIZE
        block_table = torch.arange(num_pages, dtype = torch.int32).view(cache_bsz, cache_max_seq_len // PAGE_SIZE)
        if past_len is not None:
            cache_seqlens = torch.tensor([past_len], dtype = torch.int32).repeat(bsz)
            if position is None: position = past_len
        else:
            if positions is None and position_ids is None: positions = cache_seqlens
        if position is None: position = 0
        params["block_table"] = block_table
        params["cache_seqlens"] = cache_seqlens
        params["position"] = position
        params["positions"] = positions
        params["position_ids"] = position_ids

    elif "block_table" in params:
        positions = params.get("positions")
        position_ids = params.get("position_ids")
        cache_seqlens = params.get("cache_seqlens")
        if positions is None and position_ids is None: positions = cache_seqlens
        params["cache_seqlens"] = cache_seqlens
        params["positions"] = positions
        params["position_ids"] = position_ids

    return input_ids


def prepare_flashinfer(input_ids: torch.Tensor, params: dict) -> torch.Tensor:
    bsz, seq_len = input_ids.shape

    if "batch_shape" in params:
        cache = params["cache"]
        cache_bsz, cache_max_seq_len = params["batch_shape"]
        past_len = params.get("past_len")
        cache_seqlens = params.get("cache_seqlens") if past_len is None else None
        position = params.get("position") if past_len is None else None
        positions = params.get("positions") if past_len is None else None
        position_ids = params.get("position_ids") if past_len is None else None
        assert cache_bsz >= bsz, "batch size too large for cache"
        assert cache_max_seq_len % PAGE_SIZE == 0, f"cache seq len must be a multiple of {PAGE_SIZE}"
        assert bsz * cache_max_seq_len <= cache.max_num_tokens, "Cache too small for batch shape"
        cache_bsz = min(bsz, cache_bsz)
        num_pages = cache_bsz * cache_max_seq_len // PAGE_SIZE
        block_table = torch.arange(num_pages, dtype = torch.int32).view(cache_bsz, cache_max_seq_len // PAGE_SIZE)
        if past_len is not None:
            cache_seqlens = torch.tensor([past_len], dtype = torch.int32).repeat(bsz)
            if position is None: position = past_len
        else:
            if positions is None and position_ids is None: positions = cache_seqlens
        if position is None: position = 0
        params["block_table"] = block_table
        params["cache_seqlens"] = cache_seqlens
        params["position"] = position
        params["positions"] = positions
        params["position_ids"] = position_ids

    elif "block_table" in params:
        positions = params.get("positions")
        position_ids = params.get("position_ids")
        cache_seqlens = params.get("cache_seqlens")
        if positions is None and position_ids is None: positions = cache_seqlens
        params["cache_seqlens"] = cache_seqlens
        params["positions"] = positions
        params["position_ids"] = position_ids

    # Precompute flashinfer metadata once per forward call and reuse it in
    # every attention layer to reduce Python-side decode overhead.
    block_table = params.get("block_table")
    cache_seqlens = params.get("cache_seqlens")
    if block_table is not None and cache_seqlens is not None:
        cache_seqlens = cache_seqlens.to(dtype = torch.int32).contiguous()
        block_table = block_table.to(dtype = torch.int32).contiguous()
        token_offsets = torch.arange(seq_len, dtype = torch.int32, device = cache_seqlens.device)
        batch_ids = (
            torch.arange(bsz, dtype = torch.int32, device = cache_seqlens.device)
            .view(-1, 1)
            .expand(-1, seq_len)
            .reshape(-1)
            .contiguous()
        )
        token_positions = (
            cache_seqlens.view(-1, 1)
            + token_offsets.view(1, -1)
        ).reshape(-1).contiguous()
        kv_lens = cache_seqlens + seq_len
        kv_indptr, kv_indices, kv_last_page_len = make_paged_kv_metadata(block_table, kv_lens)
        qo_indptr = torch.arange(
            0,
            (bsz + 1) * seq_len,
            seq_len,
            dtype = torch.int32,
            device = cache_seqlens.device
        )
        params["flashinfer_batch_ids"] = batch_ids
        params["flashinfer_token_positions"] = token_positions
        params["flashinfer_kv_indptr"] = kv_indptr
        params["flashinfer_kv_indices"] = kv_indices
        params["flashinfer_kv_last_page_len"] = kv_last_page_len
        params["flashinfer_qo_indptr"] = qo_indptr
        params["flashinfer_kv_lens"] = kv_lens.contiguous()

    return input_ids


def prepare_for_attn(input_ids: torch.Tensor, params: dict) -> torch.Tensor:
    """
    Add attn parameters to state
    """
    attn_mode = _resolve_attn_mode(params.get("attn_mode", "auto_nc"), params)
    params["_resolved_attn_mode"] = attn_mode
    match attn_mode:
        case "sdpa_nc":
            return prepare_sdpa_nc(input_ids, params)
        case "flash_attn":
            return prepare_flash_attn(input_ids, params)
        case "flashinfer":
            return prepare_flashinfer(input_ids, params)
        case "flash_attn_nc":
            return prepare_flash_attn_nc(input_ids, params)
        case "flashinfer_nc":
            return prepare_flashinfer_nc(input_ids, params)
        case _:
            raise ValueError(f"Unknown attn_mode: {attn_mode}")


class Attention(Module):

    def __init__(
        self,
        config: Config | None,
        key: str,
        layer_idx: int,
        hidden_size: int,
        head_dim: int,
        num_q_heads: int,
        num_kv_heads: int,
        rope_settings: RopeSettings | None,
        sm_scale: float | None = None,
        key_q: str | None = None,
        key_k: str | None = None,
        key_v: str | None = None,
        key_o: str | None = None,
        key_fused_qkv: str | None = None,
        qmap: str | None = None,
        out_dtype: torch.dtype | None = None,
        sliding_window: int  = -1,
        logit_softcapping: float = 0.0,
        q_norm: RMSNorm | LayerNorm | None = None,
        k_norm: RMSNorm | LayerNorm | None = None,
        q_proj: Linear | Module | None = None,
        k_proj: Linear | Module | None = None,
        v_proj: Linear | Module | None = None,
        kv_proj: Linear | Module | None = None,
        o_proj: Linear | Module | None = None,
        interleaved_gate: bool = False,
        use_cu_seqlens: bool = False,
        post_rope_norm: bool = False
    ):
        super().__init__(config, key, None)

        self.layer_idx = layer_idx
        self.hidden_size = hidden_size
        self.head_dim = head_dim
        self.num_q_heads = num_q_heads
        self.num_kv_heads = num_kv_heads
        self.gqa = (num_q_heads != num_kv_heads)
        self.sm_scale = sm_scale
        self.rope_settings = rope_settings
        self.rope = None
        self.out_dtype = out_dtype
        self.sliding_window = sliding_window
        self.logit_softcapping = logit_softcapping
        self.interleaved_gate = interleaved_gate
        self.use_cu_seqlens = use_cu_seqlens
        self.post_rope_norm = post_rope_norm

        if post_rope_norm:
            assert q_norm is None and k_norm is None, \
                "Post-RoPE norm only supported without weights"

        if self.num_kv_heads == 0:
            return

        if key_fused_qkv:
            assert not interleaved_gate, "Attn: interleaved_gate not implemented for fused QKV tensor"
            fkey = f"{key}.{key_fused_qkv}"
            frange_q = (0, num_q_heads * head_dim)
            frange_k = (frange_q[1], frange_q[1] + num_kv_heads * head_dim)
            frange_v = (frange_k[1], frange_k[1] + num_kv_heads * head_dim)
        else:
            fkey, frange_q, frange_k, frange_v = None, None, None, None

        if key_q or frange_q:
            f = 2 if interleaved_gate else 1
            self.q_proj = Linear(config, f"{key}.{key_q}", hidden_size, num_q_heads * head_dim * f, qmap = qmap + ".input", fkey = fkey, frange = frange_q, qbits_mod_key = "q")
            self.register_submodule(self.q_proj)
        else:
            assert q_proj
            self.q_proj = q_proj
            self.register_submodule(self.q_proj)

        if key_k or frange_k:
            assert key_v or frange_v
            self.k_proj = Linear(config, f"{key}.{key_k}", hidden_size, num_kv_heads * head_dim, qmap =  qmap + ".input", fkey = fkey, frange = frange_k, qbits_mod_key = "k")
            self.v_proj = Linear(config, f"{key}.{key_v}", hidden_size, num_kv_heads * head_dim, qmap =  qmap + ".input", fkey = fkey, frange = frange_v, qbits_mod_key = "v")
            self.register_submodule(self.k_proj)
            self.register_submodule(self.v_proj)
        else:
            if kv_proj:
                self.kv_proj = kv_proj
                self.register_submodule(self.kv_proj)
            else:
                assert k_proj and v_proj
                self.k_proj = k_proj
                self.v_proj = v_proj
                self.register_submodule(self.k_proj)
                self.register_submodule(self.v_proj)

        if key_o:
            self.o_proj = Linear(config, f"{key}.{key_o}", num_q_heads * head_dim, hidden_size, qmap =  qmap + ".o", out_dtype = out_dtype, qbits_mod_key = "o")
            self.register_submodule(self.o_proj)
        else:
            assert o_proj
            self.o_proj = o_proj
            self.register_submodule(self.o_proj)

        if q_norm:
            assert k_norm, "Must have both Q and K norms, or neither"
            self.q_norm = q_norm
            self.k_norm = k_norm
            self.register_submodule(self.q_norm)
            self.register_submodule(self.k_norm)
            if isinstance(q_norm, RMSNorm):
                self.norm_eps = q_norm.rms_norm_eps
                self.norm_constant_bias = q_norm.constant_bias
                assert self.norm_eps == k_norm.rms_norm_eps
            else:
                self.norm_eps = q_norm.layernorm_eps
                self.norm_constant_bias = 0.0
        else:
            self.q_norm = None
            self.k_norm = None
            self.norm_eps = 1e-6
            self.norm_constant_bias = 0.0

        self.caps.update({
            "kv_cache": True
        })

        self.cache_layers = []
        self.tp_cache_lookup = {}
        self.multi_kv = None
        self.tp_reduce = False

        self.q_norm_tensor = None
        self.k_norm_tensor = None
        self.flashinfer_prefill_wrapper = None
        self.flashinfer_decode_wrapper = None
        self.flashinfer_decode_wrapper_failed = False

        self.has_split_cache = False


    @override
    def optimizer_targets(self):
        q = self.q_proj.optimizer_targets()
        k = self.k_proj.optimizer_targets()
        v = self.v_proj.optimizer_targets()
        o = self.o_proj.optimizer_targets()
        return [[q, k + v, o]]


    def load_local(self, device, **kwargs):

        if self.num_kv_heads == 0:
            return

        # Cache
        for cl in self.cache_layers:
            cl.alloc(device)

        if self.rope_settings:
            self.rope = RoPE(
                device,
                self.rope_settings,
            )

        # Test if K and V proj can be fused
        if (
            device != torch.device("cpu") and
            self.k_proj.quant_type == "exl3" and
            self.v_proj.quant_type == "exl3" and
            self.k_proj.out_features == self.v_proj.out_features and
            self.k_proj.inner.K == self.v_proj.inner.K and
            self.k_proj.inner.bias is None and
            self.v_proj.inner.bias is None
        ):
            self.multi_kv = MultiLinear(self. device, [self.k_proj, self.v_proj])

        # Head norm
        if self.q_norm and isinstance(self.q_norm, RMSNorm) and not self.q_norm.span_heads:
            self.q_norm_tensor = self.q_norm.weight.data
            self.k_norm_tensor = self.k_norm.weight.data


    @override
    def load(self, device: torch.Device, **kwargs):
        super().load(device)
        self.load_local(device, **kwargs)


    @override
    def unload(self):
        super().unload()

        for cl in self.cache_layers:
            cl.free()

        self.rope = None

        if self.multi_kv is not None:
            self.multi_kv.unload()
            self.multi_kv = None

        self.q_norm_tensor = None
        self.k_norm_tensor = None
        self.flashinfer_prefill_wrapper = None
        self.flashinfer_decode_wrapper = None
        self.flashinfer_decode_wrapper_failed = False


    @override
    def forward(
        self,
        x: torch.Tensor,
        params: dict,
        out_dtype: torch.dtype | None = None
    ) -> torch.Tensor:

        if self.num_kv_heads == 0:
            x = torch.zeros_like(x, dtype = self.out_dtype)
            if self.tp_reduce:
                params["backend"].all_reduce(x, False)
        else:
            bsz, seqlen, _ = x.shape
            attn_mode = params.get("_resolved_attn_mode")
            if attn_mode is None:
                attn_mode = _resolve_attn_mode(params.get("attn_mode", "auto_nc"), params)
                params["_resolved_attn_mode"] = attn_mode
            match attn_mode:
                case "sdpa_nc":
                    x = self.decode_sdpa_nc(x, bsz, seqlen, params)
                case "flash_attn":
                    x = self.decode_flash_attn(x, bsz, seqlen, params)
                case "flashinfer":
                    x = self.decode_flashinfer(x, bsz, seqlen, params)
                case "flash_attn_nc":
                    x = self.decode_flash_attn_nc(x, bsz, seqlen, params)
                case "flashinfer_nc":
                    x = self.decode_flashinfer_nc(x, bsz, seqlen, params)
                case _:
                    raise ValueError(f"Unknown attn_mode: {attn_mode}")
            if self.tp_reduce:
                params["backend"].all_reduce(x)

        return to2(x, out_dtype, self.out_dtype)


    def project_qkv(self, x: torch.Tensor, params: dict) -> tuple:
        bsz, q_len, dim = x.shape
        q = self.q_proj.forward(x, params)

        if self.interleaved_gate:
            q, g = torch.chunk(q.view(bsz, q_len, -1, self.head_dim * 2), 2, dim = -1)
            g = g.reshape(bsz, q_len, -1)
        else:
            g = None

        if self.multi_kv is None or bsz * q_len > 32:
            k = self.k_proj.forward(x, params)
            v = self.v_proj.forward(x, params)

        else:
            x = x.view(1, bsz * q_len, dim)
            kvh = torch.empty((2, bsz * q_len, dim), dtype = torch.half, device = x.device)
            kv = torch.empty((2, bsz * q_len, self.num_kv_heads * self.head_dim), dtype = torch.half, device = x.device)
            ext.exl3_mgemm(
                x,
                self.multi_kv.ptrs_trellis,
                kv,
                self.multi_kv.ptrs_suh,
                kvh,
                self.multi_kv.ptrs_svh,
                None,
                None,
                self.multi_kv.K,
                -1,
                self.multi_kv.mcg,
                self.multi_kv.mul1,
                -1,
                -1,
                0
            )
            k = kv[0].view(bsz, q_len, self.num_kv_heads * self.head_dim)
            v = kv[1].view(bsz, q_len, self.num_kv_heads * self.head_dim)

        return q, k, v, g


    def project_o(self, o: torch.Tensor, bsz: int, seqlen: int, params: dict) -> torch.Tensor:
        o = o.reshape(bsz, seqlen, self.num_q_heads * self.head_dim)
        x = self.o_proj.forward(o, params)
        return x


    def get_flashinfer_prefill_wrapper(self) -> flashinfer.BatchPrefillWithPagedKVCacheWrapper:
        if self.flashinfer_prefill_wrapper is None:
            self.flashinfer_prefill_wrapper = get_flashinfer_prefill_wrapper_shared(self.device)
        return self.flashinfer_prefill_wrapper


    def get_flashinfer_decode_wrapper(
        self,
        backend: str = "auto",
        use_tensor_cores: bool = False,
    ) -> flashinfer.BatchDecodeWithPagedKVCacheWrapper | None:
        if self.flashinfer_decode_wrapper_failed:
            return None
        wrapper = get_flashinfer_decode_wrapper_shared(
            self.device,
            backend = backend,
            use_tensor_cores = use_tensor_cores,
            plan_key = (
                self.num_q_heads,
                self.num_kv_heads,
                self.head_dim,
                self.sliding_window,
                self.logit_softcapping,
            ),
        )
        self.flashinfer_decode_wrapper = wrapper
        return self.flashinfer_decode_wrapper


    def append_paged_kv_cache(
        self,
        k: torch.Tensor,
        v: torch.Tensor,
        cache_k: torch.Tensor,
        cache_v: torch.Tensor,
        cache_seqlens: torch.Tensor,
        block_table: torch.Tensor,
        batch_ids: torch.Tensor | None = None,
        token_positions: torch.Tensor | None = None,
        kv_indptr: torch.Tensor | None = None,
        kv_indices: torch.Tensor | None = None,
        kv_last_page_len: torch.Tensor | None = None,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        if not has_flashinfer_backend():
            raise RuntimeError("flashinfer backend requested but flashinfer is not installed")
        bsz, q_len, _, _ = k.shape
        if batch_ids is None:
            batch_ids = (
                torch.arange(bsz, dtype = torch.int32, device = k.device)
                .view(-1, 1)
                .expand(-1, q_len)
                .reshape(-1)
                .contiguous()
            )
        if token_positions is None:
            token_offsets = torch.arange(q_len, dtype = torch.int32, device = k.device)
            token_positions = (
                cache_seqlens.to(dtype = torch.int32).view(-1, 1)
                + token_offsets.view(1, -1)
            ).reshape(-1).contiguous()
        append_k = k.view(-1, self.num_kv_heads, self.head_dim).contiguous()
        append_v = v.view(-1, self.num_kv_heads, self.head_dim).contiguous()
        if kv_indptr is None or kv_indices is None or kv_last_page_len is None:
            kv_lens = cache_seqlens.to(dtype = torch.int32) + q_len
            kv_indptr, kv_indices, kv_last_page_len = make_paged_kv_metadata(block_table, kv_lens)
        flashinfer.append_paged_kv_cache(
            append_k,
            append_v,
            batch_ids,
            token_positions,
            (cache_k, cache_v),
            kv_indices,
            kv_indptr,
            kv_last_page_len,
            kv_layout = "NHD",
        )
        return kv_indptr, kv_indices, kv_last_page_len


    def decode_sdpa_nc(
        self,
        x: torch.Tensor,
        bsz: int,
        seqlen: int,
        params: dict,
    ):
        causal = params.get("causal", True)
        position = params.get("position", 0)
        positions = get_for_device(params, "positions", self.device, None)
        position_ids = get_for_device(params, "position_ids", self.device, None)
        inv_freq = get_for_device(params, "inv_freq", self.device, None)

        q, k, v, g = self.project_qkv(x, params)
        q = q.view(bsz, seqlen, self.num_q_heads, self.head_dim)
        k = k.view(bsz, seqlen, self.num_kv_heads, self.head_dim)
        v = v.view(bsz, seqlen, self.num_kv_heads, self.head_dim)

        assert self.sliding_window < 0, \
            "Torch SDPA does not support sliding window attention (SWA)"
        assert self.logit_softcapping == 0.0, \
            "Torch SDPA does not support logit softcapping"

        if self.q_norm and (not self.rope or self.q_norm_tensor is None):
            q = self.q_norm.forward(q, params, out_dtype = torch.half)
            k = self.k_norm.forward(k, params, out_dtype = torch.half)

        if self.rope:
            q, k = self.rope.apply(
                q, k,
                position,
                positions,
                position_ids,
                True,
                self.q_norm_tensor,
                self.k_norm_tensor,
                self.norm_eps,
                self.norm_constant_bias,
                inv_freq,
                self.post_rope_norm
            )

        q = q.transpose(1, 2)
        k = k.transpose(1, 2)
        v = v.transpose(1, 2)
        o = F.scaled_dot_product_attention(q, k, v, is_causal = causal, enable_gqa = self.gqa, scale = self.sm_scale)
        o = o.transpose(1, 2)
        if self.interleaved_gate: o *= g.sigmoid()

        o = self.project_o(o, bsz, seqlen, params)
        return o


    def decode_flash_attn_nc(
        self,
        x: torch.Tensor,
        bsz: int,
        seqlen: int,
        params: dict,
    ):
        if not has_flash_attn_backend():
            raise RuntimeError("flash-attn backend requested but flash_attn is not installed")

        causal = params.get("causal", True)
        position = params.get("position", 0)
        positions = get_for_device(params, "positions", self.device, None)
        position_ids = get_for_device(params, "position_ids", self.device, None)
        inv_freq = get_for_device(params, "inv_freq", self.device, None)

        q, k, v, g = self.project_qkv(x, params)
        q = q.view(bsz, seqlen, self.num_q_heads, self.head_dim)
        k = k.view(bsz, seqlen, self.num_kv_heads, self.head_dim)
        v = v.view(bsz, seqlen, self.num_kv_heads, self.head_dim)

        if self.q_norm and (not self.rope or self.q_norm_tensor is None):
            q = self.q_norm.forward(q, params, out_dtype = torch.half)
            k = self.k_norm.forward(k, params, out_dtype = torch.half)

        if self.rope:
            q, k = self.rope.apply(
                q, k,
                position,
                positions,
                position_ids,
                True,
                self.q_norm_tensor,
                self.k_norm_tensor,
                self.norm_eps,
                self.norm_constant_bias,
                inv_freq,
                self.post_rope_norm
            )

        if self.use_cu_seqlens and (cu_seqlens := get_for_device(params, "cu_seqlens", self.device, None)) is not None:
            max_seqlen = params["max_seqlen"]
            o = flash_attn_varlen_func(
                q = q.squeeze(0),
                k = k.squeeze(0),
                v = v.squeeze(0),
                cu_seqlens_q = cu_seqlens,
                cu_seqlens_k = cu_seqlens,
                max_seqlen_q = max_seqlen,
                max_seqlen_k = max_seqlen,
                causal = causal,
                softmax_scale = self.sm_scale,
                window_size = (self.sliding_window, self.sliding_window),
                softcap = self.logit_softcapping
            )
        else:
            o = flash_attn_func(
                q = q,
                k = k,
                v = v,
                causal = causal,
                softmax_scale = self.sm_scale,
                window_size = (self.sliding_window, self.sliding_window),
                softcap = self.logit_softcapping
            )

        o = o.view((bsz, seqlen, self.num_q_heads * self.head_dim))
        if self.interleaved_gate: o *= g.sigmoid()

        o = self.project_o(o, bsz, seqlen, params)
        return o


    def decode_flashinfer_nc(
        self,
        x: torch.Tensor,
        bsz: int,
        seqlen: int,
        params: dict,
    ):
        if not has_flashinfer_backend():
            raise RuntimeError("flashinfer backend requested but flashinfer is not installed")
        causal = params.get("causal", True)
        position = params.get("position", 0)
        positions = get_for_device(params, "positions", self.device, None)
        position_ids = get_for_device(params, "position_ids", self.device, None)
        inv_freq = get_for_device(params, "inv_freq", self.device, None)

        q, k, v, g = self.project_qkv(x, params)
        q = q.view(bsz, seqlen, self.num_q_heads, self.head_dim)
        k = k.view(bsz, seqlen, self.num_kv_heads, self.head_dim)
        v = v.view(bsz, seqlen, self.num_kv_heads, self.head_dim)

        if self.q_norm and (not self.rope or self.q_norm_tensor is None):
            q = self.q_norm.forward(q, params, out_dtype = torch.half)
            k = self.k_norm.forward(k, params, out_dtype = torch.half)

        if self.rope:
            q, k = self.rope.apply(
                q, k,
                position,
                positions,
                position_ids,
                True,
                self.q_norm_tensor,
                self.k_norm_tensor,
                self.norm_eps,
                self.norm_constant_bias,
                inv_freq,
                self.post_rope_norm
            )

        def _sdpa_nc_fallback() -> torch.Tensor:
            assert self.sliding_window < 0, \
                "Torch SDPA fallback does not support sliding window attention (SWA)"
            assert self.logit_softcapping == 0.0, \
                "Torch SDPA fallback does not support logit softcapping"

            if self.use_cu_seqlens and (cu_seqlens := get_for_device(params, "cu_seqlens", self.device, None)) is not None:
                q_ = q.squeeze(0)
                k_ = k.squeeze(0)
                v_ = v.squeeze(0)
                out_segs = []
                for i in range(cu_seqlens.numel() - 1):
                    a = cu_seqlens[i].item()
                    b = cu_seqlens[i + 1].item()
                    out_seg = F.scaled_dot_product_attention(
                        q_[a:b].transpose(0, 1).unsqueeze(0),
                        k_[a:b].transpose(0, 1).unsqueeze(0),
                        v_[a:b].transpose(0, 1).unsqueeze(0),
                        is_causal = causal,
                        enable_gqa = self.gqa,
                        scale = self.sm_scale,
                    )
                    out_segs.append(out_seg.squeeze(0).transpose(0, 1))
                return torch.cat(out_segs, dim = 0).unsqueeze(0)

            return F.scaled_dot_product_attention(
                q.transpose(1, 2),
                k.transpose(1, 2),
                v.transpose(1, 2),
                is_causal = causal,
                enable_gqa = self.gqa,
                scale = self.sm_scale,
            ).transpose(1, 2)

        if not causal:
            o = _sdpa_nc_fallback()
            o = o.view((bsz, seqlen, self.num_q_heads * self.head_dim))
            if self.interleaved_gate: o *= g.sigmoid()
            o = self.project_o(o, bsz, seqlen, params)
            return o

        window_left = self.sliding_window if self.sliding_window >= 0 else -1
        logits_soft_cap = self.logit_softcapping if self.logit_softcapping > 0 else None

        try:
            if self.use_cu_seqlens and (cu_seqlens := get_for_device(params, "cu_seqlens", self.device, None)) is not None:
                q_ = q.squeeze(0)
                k_ = k.squeeze(0)
                v_ = v.squeeze(0)
                out_segs = []
                for i in range(cu_seqlens.numel() - 1):
                    a = cu_seqlens[i].item()
                    b = cu_seqlens[i + 1].item()
                    out_segs.append(
                        flashinfer.single_prefill_with_kv_cache(
                            q_[a:b],
                            k_[a:b],
                            v_[a:b],
                            causal = causal,
                            kv_layout = "NHD",
                            pos_encoding_mode = "NONE",
                            sm_scale = self.sm_scale,
                            window_left = window_left,
                            logits_soft_cap = logits_soft_cap,
                        )
                    )
                o = torch.cat(out_segs, dim = 0).unsqueeze(0)
            else:
                o = torch.stack([
                    flashinfer.single_prefill_with_kv_cache(
                        q[i],
                        k[i],
                        v[i],
                        causal = causal,
                        kv_layout = "NHD",
                        pos_encoding_mode = "NONE",
                        sm_scale = self.sm_scale,
                        window_left = window_left,
                        logits_soft_cap = logits_soft_cap,
                    )
                    for i in range(bsz)
                ], dim = 0)
        except RuntimeError as ex:
            if "Unsupported head_dim" not in str(ex):
                raise
            o = _sdpa_nc_fallback()

        o = o.view((bsz, seqlen, self.num_q_heads * self.head_dim))
        if self.interleaved_gate: o *= g.sigmoid()

        o = self.project_o(o, bsz, seqlen, params)
        return o


    def decode_flash_attn(
        self,
        x: torch.Tensor,
        bsz: int,
        seqlen: int,
        params: dict,
    ):
        if not has_flash_attn_backend():
            raise RuntimeError("flash-attn backend requested but flash_attn is not installed")

        cache = params.get("cache")
        block_table = get_for_device(params, "block_table", self.device)
        cache_seqlens = get_for_device(params, "cache_seqlens", self.device)
        position = params.get("position", 0)
        positions = get_for_device(params, "positions", self.device, None)
        position_ids = get_for_device(params, "position_ids", self.device, None)
        inv_freq = get_for_device(params, "inv_freq", self.device, None)
        causal = params.get("causal", True)

        q, k, v, g = self.project_qkv(x, params)
        q = q.view(bsz, seqlen, self.num_q_heads, self.head_dim)
        k = k.view(bsz, seqlen, self.num_kv_heads, self.head_dim)
        v = v.view(bsz, seqlen, self.num_kv_heads, self.head_dim)

        if self.q_norm and (not self.rope or self.q_norm_tensor is None):
            q = self.q_norm.forward(q, params, out_dtype = torch.half)
            k = self.k_norm.forward(k, params, out_dtype = torch.half)

        if self.rope:
            q, k = self.rope.apply(
                q, k,
                position,
                positions,
                position_ids,
                True,
                self.q_norm_tensor,
                self.k_norm_tensor,
                self.norm_eps,
                self.norm_constant_bias,
                inv_freq,
                self.post_rope_norm
            )

        if self.has_split_cache:
            cache_k, cache_v = self.tp_cache_lookup[cache].get_kv(cache_seqlens, block_table)
        else:
            cache_k, cache_v = cache.get_layer(self.layer_idx, cache_seqlens, block_table)

        o = flash_attn_with_kvcache(
            q = q,
            k = k,
            v = v,
            k_cache = cache_k,
            v_cache = cache_v,
            block_table = block_table,
            cache_seqlens = cache_seqlens,
            causal = causal,
            softmax_scale = self.sm_scale,
            window_size = (self.sliding_window, self.sliding_window),
            softcap = self.logit_softcapping
        )

        if self.has_split_cache:
            self.tp_cache_lookup[cache].update_kv(cache_seqlens, block_table, cache_k, cache_v, seqlen)
        else:
            cache.update_layer(self.layer_idx, cache_seqlens, block_table, cache_k, cache_v, seqlen)

        o = o.view((bsz, seqlen, self.num_q_heads * self.head_dim))
        if self.interleaved_gate: o *= g.sigmoid()

        o = self.project_o(o, bsz, seqlen, params)
        return o


    def decode_flashinfer(
        self,
        x: torch.Tensor,
        bsz: int,
        seqlen: int,
        params: dict,
    ):
        if not has_flashinfer_backend():
            raise RuntimeError("flashinfer backend requested but flashinfer is not installed")
        cache = params.get("cache")
        block_table = get_for_device(params, "block_table", self.device)
        cache_seqlens = get_for_device(params, "cache_seqlens", self.device)
        batch_ids = get_for_device(params, "flashinfer_batch_ids", self.device, None)
        token_positions = get_for_device(params, "flashinfer_token_positions", self.device, None)
        kv_indptr = get_for_device(params, "flashinfer_kv_indptr", self.device, None)
        kv_indices = get_for_device(params, "flashinfer_kv_indices", self.device, None)
        kv_last_page_len = get_for_device(params, "flashinfer_kv_last_page_len", self.device, None)
        qo_indptr = get_for_device(params, "flashinfer_qo_indptr", self.device, None)
        position = params.get("position", 0)
        positions = get_for_device(params, "positions", self.device, None)
        position_ids = get_for_device(params, "position_ids", self.device, None)
        inv_freq = get_for_device(params, "inv_freq", self.device, None)
        causal = params.get("causal", True)

        q, k, v, g = self.project_qkv(x, params)
        q = q.view(bsz, seqlen, self.num_q_heads, self.head_dim)
        k = k.view(bsz, seqlen, self.num_kv_heads, self.head_dim)
        v = v.view(bsz, seqlen, self.num_kv_heads, self.head_dim)

        # TODO: Add LayerNorm option to fused norm/RoPE kernel
        if self.q_norm and (not self.rope or self.q_norm_tensor is None):
            q = self.q_norm.forward(q, params, out_dtype = torch.half)
            k = self.k_norm.forward(k, params, out_dtype = torch.half)

        if self.rope:
            q, k = self.rope.apply(
                q, k,
                position,
                positions,
                position_ids,
                True,
                self.q_norm_tensor,
                self.k_norm_tensor,
                self.norm_eps,
                self.norm_constant_bias,
                inv_freq,
                self.post_rope_norm
            )

        if self.has_split_cache:
            cache_k, cache_v = self.tp_cache_lookup[cache].get_kv(cache_seqlens, block_table)
        else:
            cache_k, cache_v = cache.get_layer(self.layer_idx, cache_seqlens, block_table)

        kv_indptr, kv_indices, kv_last_page_len = self.append_paged_kv_cache(
            k,
            v,
            cache_k,
            cache_v,
            cache_seqlens,
            block_table,
            batch_ids = batch_ids,
            token_positions = token_positions,
            kv_indptr = kv_indptr,
            kv_indices = kv_indices,
            kv_last_page_len = kv_last_page_len,
        )
        if qo_indptr is None:
            qo_indptr = torch.arange(
                0,
                (bsz + 1) * seqlen,
                seqlen,
                dtype = torch.int32,
                device = self.device
            )
        window_left = self.sliding_window if self.sliding_window >= 0 else -1
        logits_soft_cap = self.logit_softcapping if self.logit_softcapping > 0 else None
        q_flat = q.view(-1, self.num_q_heads, self.head_dim).contiguous()
        o = None

        # Single-token decode is the hot path for generation. Prefer flashinfer's
        # decode wrapper here and fall back to the prefill wrapper on any runtime
        # incompatibility to preserve correctness.
        use_decode_wrapper = params.get("flashinfer_use_decode_wrapper", True)
        decode_backend = params.get("flashinfer_decode_backend", "auto")
        decode_use_tensor_cores = params.get("flashinfer_decode_use_tensor_cores", False)
        decode_fixed_split_size = params.get("flashinfer_decode_fixed_split_size", None)
        decode_disable_split_kv = params.get("flashinfer_decode_disable_split_kv", False)
        if use_decode_wrapper and causal and seqlen == 1:
            decode_wrapper = self.get_flashinfer_decode_wrapper(
                backend = decode_backend,
                use_tensor_cores = decode_use_tensor_cores,
            )
            if decode_wrapper is not None:
                try:
                    wrapper_id = id(decode_wrapper)
                    if decode_wrapper.use_tensor_cores:
                        plan_marker = (
                            id(decode_wrapper),
                            kv_indptr.data_ptr(),
                            kv_indices.data_ptr(),
                            kv_last_page_len.data_ptr(),
                            self.num_q_heads,
                            self.num_kv_heads,
                            self.head_dim,
                            window_left,
                            logits_soft_cap,
                            q.dtype,
                            cache_k.dtype,
                        )
                        planned = params.get("flashinfer_decode_planned_wrappers")
                        if planned is None:
                            planned = {}
                            params["flashinfer_decode_planned_wrappers"] = planned
                        should_replan = planned.get(wrapper_id) != plan_marker
                    else:
                        plan_marker = (
                            kv_indptr.numel(),
                            kv_indices.numel(),
                            kv_last_page_len.numel(),
                            self.num_q_heads,
                            self.num_kv_heads,
                            self.head_dim,
                            window_left,
                            logits_soft_cap,
                            q.dtype,
                            cache_k.dtype,
                        )
                        should_replan = (
                            getattr(decode_wrapper, "_exllamav3_decode_plan_marker", None)
                            != plan_marker
                        )
                    if (
                        not should_replan
                        and can_reuse_flashinfer_decode_plan(
                            decode_wrapper,
                            kv_indptr,
                            kv_indices,
                            kv_last_page_len,
                        )
                    ):
                        decode_wrapper._paged_kv_indices_buf.copy_(
                            kv_indices,
                            non_blocking = (
                                kv_indices.device
                                == decode_wrapper._paged_kv_indices_buf.device
                            ),
                        )
                        decode_wrapper._paged_kv_last_page_len_buf.copy_(
                            kv_last_page_len,
                            non_blocking = (
                                kv_last_page_len.device
                                == decode_wrapper._paged_kv_last_page_len_buf.device
                            ),
                        )
                        if _FLASHINFER_RUNTIME_STATS_ENABLED:
                            _flashinfer_runtime_stat("decode_plan_reuses")
                    else:
                        if _FLASHINFER_RUNTIME_STATS_ENABLED:
                            _flashinfer_runtime_stat("decode_plan_calls")
                        decode_wrapper.plan(
                            indptr = kv_indptr,
                            indices = kv_indices,
                            last_page_len = kv_last_page_len,
                            num_qo_heads = self.num_q_heads,
                            num_kv_heads = self.num_kv_heads,
                            head_dim = self.head_dim,
                            page_size = PAGE_SIZE,
                            pos_encoding_mode = "NONE",
                            sm_scale = self.sm_scale,
                            window_left = window_left,
                            logits_soft_cap = logits_soft_cap,
                            q_data_type = q.dtype,
                            kv_data_type = cache_k.dtype,
                            o_data_type = q.dtype,
                            fixed_split_size = decode_fixed_split_size,
                            disable_split_kv = decode_disable_split_kv,
                        )
                        if decode_wrapper.use_tensor_cores:
                            planned[wrapper_id] = plan_marker
                        else:
                            decode_wrapper._exllamav3_decode_plan_marker = plan_marker
                    o = decode_wrapper.run(
                        q_flat,
                        (cache_k, cache_v),
                        q_len_per_req = 1,
                    ).view(bsz, seqlen, self.num_q_heads, self.head_dim)
                except (RuntimeError, TypeError, ValueError, AssertionError):
                    self.flashinfer_decode_wrapper_failed = True
                    self.flashinfer_decode_wrapper = None

        if o is None:
            retried_workspace_growth = False
            while True:
                prefill_wrapper = self.get_flashinfer_prefill_wrapper()
                try:
                    planned = params.get("flashinfer_prefill_planned_wrappers")
                    if planned is None:
                        planned = {}
                        params["flashinfer_prefill_planned_wrappers"] = planned
                    plan_marker = (
                        qo_indptr.data_ptr(),
                        kv_indptr.data_ptr(),
                        kv_indices.data_ptr(),
                        kv_last_page_len.data_ptr(),
                        self.num_q_heads,
                        self.num_kv_heads,
                        self.head_dim,
                        causal,
                        window_left,
                        logits_soft_cap,
                        q.dtype,
                        cache_k.dtype,
                    )
                    wrapper_id = id(prefill_wrapper)
                    if planned.get(wrapper_id) != plan_marker:
                        if _FLASHINFER_RUNTIME_STATS_ENABLED:
                            _flashinfer_runtime_stat("prefill_plan_calls")
                        prefill_wrapper.plan(
                            qo_indptr = qo_indptr,
                            paged_kv_indptr = kv_indptr,
                            paged_kv_indices = kv_indices,
                            paged_kv_last_page_len = kv_last_page_len,
                            num_qo_heads = self.num_q_heads,
                            num_kv_heads = self.num_kv_heads,
                            head_dim_qk = self.head_dim,
                            page_size = PAGE_SIZE,
                            causal = causal,
                            pos_encoding_mode = "NONE",
                            sm_scale = self.sm_scale,
                            window_left = window_left,
                            logits_soft_cap = logits_soft_cap,
                            q_data_type = q.dtype,
                            kv_data_type = cache_k.dtype,
                            o_data_type = q.dtype,
                        )
                        planned[wrapper_id] = plan_marker
                    else:
                        if _FLASHINFER_RUNTIME_STATS_ENABLED:
                            _flashinfer_runtime_stat("prefill_plan_reuses")
                    o = prefill_wrapper.run(
                        q_flat,
                        (cache_k, cache_v),
                    ).view(bsz, seqlen, self.num_q_heads, self.head_dim)
                    break
                except RuntimeError as ex:
                    if (
                        not retried_workspace_growth
                        and grow_flashinfer_workspace_for_exception(self.device, ex)
                    ):
                        retried_workspace_growth = True
                        self.flashinfer_prefill_wrapper = None
                        self.flashinfer_decode_wrapper = None
                        self.flashinfer_decode_wrapper_failed = False
                        continue
                    raise

        if self.has_split_cache:
            self.tp_cache_lookup[cache].update_kv(cache_seqlens, block_table, cache_k, cache_v, seqlen)
        else:
            cache.update_layer(self.layer_idx, cache_seqlens, block_table, cache_k, cache_v, seqlen)

        o = o.view((bsz, seqlen, self.num_q_heads * self.head_dim))
        if self.interleaved_gate: o *= g.sigmoid()

        o = self.project_o(o, bsz, seqlen, params)
        return o


    def make_tp_allocation(self, options: dict) -> list[TPAllocation]:
        storage = 0
        storage += self.q_proj.storage_size()
        storage += self.k_proj.storage_size()
        storage += self.v_proj.storage_size()
        storage += self.o_proj.storage_size()
        for cl in self.cache_layers:
            storage += cl.storage_size()
        overhead_d = 0
        overhead_d += self.hidden_size * (self.out_dtype or torch.half).itemsize
        overhead_s = 0
        for cl in self.cache_layers:
            overhead_s += cl.overhead_size()
        overhead_s += 2 * self.num_q_heads * self.head_dim * torch.half.itemsize  # q, o
        overhead_s += 2 * self.num_kv_heads * self.head_dim * torch.half.itemsize  # k, v
        recons = max(
            self.q_proj.recons_size(),
            self.k_proj.recons_size(),
            self.v_proj.recons_size(),
            self.o_proj.recons_size(),
        )
        channel_width = 1
        channels_to_split = self.num_kv_heads
        while channel_width * self.head_dim < 128:
            assert channels_to_split % 2 == 0, \
                "Model's K/V heads cannot divide into 128-channel tensors"
            channel_width *= 2
            channels_to_split //= 2
        assert (channel_width * self.head_dim) % 128 == 0, \
            "Model's K/V heads cannot divide into 128-channel tensors"
        # TODO: Account for flashinfer temp VRAM usage
        tpa = TPAllocation(
            key = self.key,
            channel_width = channel_width,
            channel_unit = "heads",
            storage_per_device = 0,
            storage_to_split = storage,
            overhead_per_device = overhead_d,
            overhead_to_split = overhead_s,
            recons_temp = recons,
            channels_to_split = channels_to_split,
            limit_key = "attn"
        )
        return [tpa]


    def tp_export(self, plan, producer):
        assert self.device is not None, "Cannot export module for TP before loading."

        def _export(child):
            nonlocal producer
            return child.tp_export(plan, producer) if child is not None else None

        return {
            "cls": Attention,
            "kwargs": {
                "key": self.key,
                "layer_idx": self.layer_idx,
                "hidden_size": self.hidden_size,
                "head_dim": self.head_dim,
                "rope_settings": self.rope_settings,
                "sm_scale": self.sm_scale,
                "out_dtype": self.out_dtype,
                "sliding_window": self.sliding_window,
                "logit_softcapping": self.logit_softcapping,
                "post_rope_norm": self.post_rope_norm,
            },
            "num_kv_heads": self.num_kv_heads,
            **{name: _export(getattr(self, name, None)) for name in (
                "q_norm",
                "k_norm",
                "q_proj",
                "k_proj",
                "v_proj",
                "kv_proj",
                "o_proj",
            )},
            "device": self.device,
            "cache_layers": [
                cl.tp_export(plan) for cl in self.cache_layers
            ],
            "n_gqa": self.num_q_heads // self.num_kv_heads
        }


    @staticmethod
    def tp_import(local_context, exported, plan, **kwargs):
        key = exported["kwargs"]["key"]
        head_dim = exported["kwargs"]["head_dim"]
        n_gqa = exported["n_gqa"]
        device = local_context["device"]
        first, last, unit = plan[key]
        assert unit == "heads"
        num_kv_heads = last - first
        num_q_heads = num_kv_heads * n_gqa

        q_split = (True, first * head_dim * n_gqa, last * head_dim * n_gqa) \
            if num_kv_heads else None
        kv_split = (True, first * head_dim, last * head_dim) \
            if num_kv_heads else None
        o_split = (False, first * head_dim * n_gqa, last * head_dim * n_gqa) \
            if num_kv_heads else None
        norm_q_split = (first * n_gqa, last * n_gqa) \
            if num_kv_heads else None
        norm_k_split = (first, last) \
            if num_kv_heads else None

        # def _import(name):
        #     nonlocal exported, plan
        #     return exported[name]["cls"].tp_import(local_context, exported[name], plan) \
        #         if exported.get(name) else None

        def _import_split(name, split):
            nonlocal exported, plan
            return exported[name]["cls"].tp_import_split(local_context, exported[name], plan, split) \
                if split and exported.get(name) else None

        module = Attention(
            config = None,
            **exported["kwargs"],
            num_q_heads = num_q_heads,
            num_kv_heads = num_kv_heads,
            q_norm = _import_split("q_norm", norm_q_split),
            k_norm = _import_split("k_norm", norm_k_split),
            q_proj = _import_split("q_proj", q_split),
            k_proj = _import_split("k_proj", kv_split),
            v_proj = _import_split("v_proj", kv_split),
            kv_proj = _import_split("kv_proj", kv_split),
            o_proj = _import_split("o_proj", o_split),
        )

        if num_kv_heads:
            cache_layers = exported["cache_layers"]
            if len(cache_layers):
                module.has_split_cache = True
                for cl in exported["cache_layers"]:
                    cli = cl["cls"](None, module, **cl["args"])
                    module.cache_layers.append(cli)
                    module.tp_cache_lookup[cl["args"]["cache_id"]] = cli

        module.device = device
        if not kwargs.get("skip_reduction"):
            module.tp_reduce = True
        module.load_local(device)
        torch.cuda.synchronize()
        return module



from __future__ import annotations
from typing import override
from .llama import LlamaConfig, LlamaModel
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .qwen2_5_vl import Qwen2_5VLConfig

# Qwen2 is identical to Llama except for bias on Q, K and V projections, but Linear module automatically
# detects *.bias tensor

class Qwen2Config(LlamaConfig):
    arch_string = "Qwen2ForCausalLM"

    def __init__(
        self,
        directory: str,
        **kwargs,
    ):
        super().__init__(
            directory,
            derived_model = {"text": Qwen2Model},
            **kwargs
        )


class Qwen2Model(LlamaModel):
    config_class = Qwen2Config

    def __init__(
        self,
        config: Qwen2Config | Qwen2_5VLConfig,
        **kwargs
    ):
        super().__init__(config, **kwargs)


    @override
    def default_chat_prompt(self, prompt: str, system_prompt: str = None) -> str:
        p = ""
        if system_prompt:
            p += f"<|im_start|>system\n"
            p += f"{system_prompt}<|im_end|>\n"
        p += f"<|im_start|>user\n"
        p += f"{prompt}<|im_end|>\n"
        p += f"<|im_start|>assistant\n"
        return p


from __future__ import annotations
import torch
import torch.nn.functional as F

def sqnr(a: torch.Tensor, b: torch.Tensor, eps: float = 1e-8):
    a_flat = a.view(a.shape[0], -1)
    b_flat = b.view(b.shape[0], -1)
    signal_power = torch.sum(b_flat ** 2, dim = 1)
    noise_power = torch.sum((a_flat - b_flat) ** 2, dim = 1) + eps
    return 10.0 * torch.log10(signal_power / noise_power).mean().item() # dB

def cosine_error(a: torch.Tensor, b: torch.Tensor, eps: float = 1e-8):
    a_flat = a.view(a.shape[0], -1)
    b_flat = b.view(b.shape[0], -1)
    cos_sim = F.cosine_similarity(a_flat, b_flat, dim = 1, eps = eps)
    return 1.0 - cos_sim.mean().item()



from __future__ import annotations
from typing import override
import torch
from torch import nn
from ..model.config import Config
from . import Module
from ..ext import exllamav3_ext as ext
from ..model.model_tp_alloc import TPAllocation

class RMSNorm(Module):

    def __init__(
        self,
        config: Config | None,
        key: str,
        rms_norm_eps: float,
        out_dtype: torch.dtype | None = None,
        qmap: str | None = None,
        constant_bias: float = 0.0,
        span_heads: bool = False,
        unweighted: bool = False
    ):
        super().__init__(config, key, None)
        assert qmap is None, "No quant scheme for RMSNorm"
        self.module_name = "RMSNorm"

        self.weight = None
        self.rms_norm_eps = rms_norm_eps
        self.out_dtype = out_dtype
        self._numel = None
        self.constant_bias = constant_bias
        self.span_heads = span_heads
        self.unweighted = unweighted

    @override
    def optimizer_targets(self):
        return []

    @override
    def load(self, device: torch.device, **kwargs):
        self.device = device
        if not self.unweighted:
            weight = self.config.stc.get_tensor(f"{self.key}.weight", self.device, float2half = True)
            self._numel = weight.numel()
            self.weight = nn.Parameter(weight, requires_grad = False)
        else:
            self._numel = 0

    @override
    def unload(self):
        self.device = None
        self.weight = None

    @override
    def get_tensors(self):
        return {} if self.unweighted else {
            f"{self.key}.weight": self.weight.data
        }

    def forward_torch(
        self,
        x: torch.Tensor,
        params: dict,
        out_dtype: torch.dtype | None = None,
    ) -> torch.Tensor:
        dtype = x.dtype
        x = x.float()
        var = x.pow(2).mean(dim = -1, keepdim = True) + self.rms_norm_eps
        x = x * torch.rsqrt(var)
        x = x.to(dtype)
        if not self.unweighted:
            x = x * self.weight if self.constant_bias == 0.0 else x * (self.weight + self.constant_bias)
        x = x.to(out_dtype or self.out_dtype)
        return x

    @override
    def weights_numel(self):
        return self._numel

    @override
    def forward(
        self,
        x: torch.Tensor,
        params,
        out_dtype: torch.dtype | None = None,
    ) -> torch.Tensor:
        dtype = out_dtype or self.out_dtype

        # ext.rms_norm expects row-major 2D inputs for the standard
        # per-channel RMSNorm path. Flattening keeps results consistent across
        # chunk/prefill shapes (e.g. [B, T, C] vs [B*T, C]).
        if not self.span_heads and x.dim() > 2:
            x_2d = x.view(-1, x.shape[-1]).contiguous()
            y_2d = torch.empty_like(x_2d, dtype = dtype)
            ext.rms_norm(
                x_2d,
                self.weight,
                y_2d,
                self.rms_norm_eps,
                self.constant_bias,
                self.span_heads,
            )
            return y_2d.view_as(x)

        y = torch.empty_like(x, dtype = dtype)
        ext.rms_norm(
            x,
            self.weight,
            y,
            self.rms_norm_eps,
            self.constant_bias,
            self.span_heads,
        )
        return y

    def make_tp_allocation(self, options: dict) -> list[TPAllocation]:
        if self.unweighted:
            return []
        stc = self.config.stc
        storage = sum(stc.get_tensor_sizes(self.key))
        overhead = storage // 2 * (self.out_dtype or torch.half).itemsize
        tpa = TPAllocation(
            key = self.key,
            storage_per_device = storage,
            overhead_per_device = overhead,
        )
        return [tpa]

    def tp_export(self, plan, producer):
        assert self.device is not None, "Cannot export module for TP before loading."
        return {
            "cls": RMSNorm,
            "kwargs": {
                "key": self.key,
                "rms_norm_eps": self.rms_norm_eps,
                "out_dtype": self.out_dtype,
                "constant_bias": self.constant_bias,
            },
            "weight": producer.send(self.weight),
            "device": self.device,
        }

    @staticmethod
    def tp_import(local_context, exported, plan):
        consumer = local_context["consumer"]
        device = local_context["device"]
        module = RMSNorm(
            config = None,
            **exported["kwargs"],
        )
        module.device = device
        w = consumer.recv(exported["weight"], cuda = True)
        module.weight = nn.Parameter(w) if w is not None else None
        torch.cuda.synchronize()
        return module

    @staticmethod
    def tp_import_split(local_context, exported, plan, split):
        consumer = local_context["consumer"]
        device = local_context["device"]
        first, last = split
        module = RMSNorm(
            config = None,
            **exported["kwargs"],
        )
        module.device = device

        w = consumer.recv(exported["weight"], cuda = True)
        if w is not None:
            if w.dim() == 2:
                w = w[first : last, :]
            module.weight = nn.Parameter(w.to(module.device).contiguous())

        return module



from __future__ import annotations
from functools import cached_property
from typing import override
import torch
import numpy as np
from ..model.config import Config
from . import Module
from .quant import LinearFP16, LinearEXL3
from .quant.exl3_lib import quantize_exl3
from ..ext import exllamav3_ext as ext
from ..conversion.allocation import allocate_linear
from ..model.model_tp_alloc import TPAllocation


class Linear(Module):

    def __init__(
        self,
        config: Config | None,
        key: str,
        in_features: int,
        out_features: int,
        qmap: str | None = None,
        alt_key: str | None = None,
        qbits_key: str = "bits",
        qbits_mod_key: str = "",
        fkey : str | None = None,
        frange: tuple[int, int] | None = None,
        fidx: int | None = None,
        caps: dict = None,
        softcap: float = 0.0,
        pad_to: int = 128,
        full_in_features: int | None = None,
        full_out_features: int | None = None,
        first_in_feature: int | None = None,
        first_out_feature: int | None = None,
        out_dtype: torch.dtype | None = None,
        allow_input_padding: bool = False,
        post_scale: float = 1.0,
        transposed_load: bool = True
    ):
        super().__init__(config, key, qmap)

        self.alt_key = alt_key
        self.in_features_unpadded = in_features
        self.in_features = (in_features + pad_to - 1) // pad_to * pad_to
        self.out_features_unpadded = out_features
        self.out_features = (out_features + pad_to - 1) // pad_to * pad_to
        self.full_in_features = full_in_features if full_in_features is not None else self.in_features
        self.full_out_features = full_out_features if full_out_features is not None else self.out_features
        self.first_in_feature = first_in_feature if first_in_feature is not None else 0
        self.first_out_feature = first_out_feature if first_out_feature is not None else 0
        self.inner = None
        self.qbits_key = qbits_key
        self.qbits_mod_key = qbits_mod_key
        self.fkey = fkey
        self.frange = frange
        self.fidx = fidx
        self.quant_type = None
        self.softcap = softcap
        self.is_sliced = self.in_features < self.full_in_features or self.out_features < self.full_out_features
        self.out_dtype = out_dtype
        self.post_scale = post_scale
        self.transposed_load = transposed_load

        assert self.in_features_unpadded == self.in_features or allow_input_padding, \
            f"Input padding is not allowed for {self.key}, in_dim: {self.in_features_unpadded}, pad_to: {pad_to}"

        if caps is not None:
            self.caps.update(caps)


    @override
    def optimizer_targets(self):
        return [self.key]


    def pad_out(self, w: torch.Tensor | None) -> torch.Tensor | None:
        if w is None or self.out_features == self.out_features_unpadded and self.in_features == self.in_features_unpadded:
            return w
        if w.dim() == 2:
            padded = torch.zeros((self.in_features, self.out_features), dtype = w.dtype, device = w.device)
            if self.out_features != self.out_features_unpadded:
                padded[:, :w.shape[1]] = w
            elif self.in_features != self.in_features_unpadded:
                padded[:w.shape[0], :] = w
        else:
            assert w.dim() == 1
            padded = torch.zeros((self.out_features,), dtype = w.dtype, device = w.device)
            padded[:w.shape[0]] = w
        return padded.contiguous()


    # Row scales (Transformers-specific?)
    def apply_fp8_scales_(self, weight: torch.Tensor, scale: torch.Tensor):
        ws = weight.shape
        ss = scale.shape
        assert len(ss) == 0 or len(ws) == len(ss) == 2
        if len(ss) == 2:
            if 1 < ss[0] < ws[0]:
                scale = torch.nn.functional.pad(scale, (0, 0, 0, ws[0] - ss[0]), "constant", 1)
            if 1 < ss[1] < ws[1]:
                scale = torch.nn.functional.pad(scale, (0, ws[1] - ss[1]), "constant", 1)
        weight = weight.float() * scale
        return weight.view(ws).half()


    def apply_fp8_scales_inv_(self, weight: torch.Tensor, scale_inv: torch.Tensor):
        ws = weight.shape
        ss = scale_inv.shape
        if len(ss) > 0:
            assert len(ws) == len(ss) == 2
            assert all(w == s * 128 for w, s in zip(ws, ss))
            weight = weight.view(ss[0], 128, ss[1], 128)
            scale_inv = scale_inv.view(ss[0], 1, ss[1], 1)
        weight = weight.float() * scale_inv
        return weight.view(ws).half()


    def load_fp16(self, key: str) -> bool:

        if self.config.stc.has_tensor_group(key, ["weight"]):

            self.used_alt_key = key == self.alt_key
            dev = "cpu" if self.is_sliced else self.device
            pad1 = (self.out_features,) if not self.is_sliced else None
            pad2 = (self.in_features, self.out_features) if not self.is_sliced else None
            scale = self.config.stc.get_tensor(key + ".weight_scale", dev, transpose = self.transposed_load, optional = True, no_defer = True)
            scale_inv = self.config.stc.get_tensor(key + ".weight_scale_inv", dev, transpose = self.transposed_load, optional = True, no_defer = True)
            assert scale is None or scale_inv is None
            no_defer = scale is not None or scale_inv is not None
            weight = self.config.stc.get_tensor(key + ".weight", dev, float2half = True, transpose = self.transposed_load, pad_to = pad2, no_defer = no_defer)
            bias = self.config.stc.get_tensor(key + ".bias", dev, float2half = True, optional = True, pad_to = pad1)
            if scale is not None:
                weight = self.apply_fp8_scales_(weight, scale)
            elif scale_inv is not None:
                weight = self.apply_fp8_scales_inv_(weight, scale_inv)
            self.inner = LinearFP16(
                self.in_features,
                self.out_features,
                weight,
                bias,
                self.full_in_features,
                self.full_out_features,
                self.first_in_feature,
                self.first_out_feature,
                self.out_dtype,
                key = self.key
            )
            if self.is_sliced:
                self.inner.swap_device = self.device
                self.inner.unswap_cpu()
            self.quant_type = "fp16"
            return True

        # Special dumb loading mode for Qwen3VLMoE
        elif self.fkey and self.config.stc.has_tensor(self.fkey) and self.fidx is not None:

            # Load and split fused weight along first dim
            weight = self.config.stc.get_tensor(
                self.fkey,
                self.device,
                transpose = self.transposed_load,
                no_defer = True,
                fidx = self.fidx
            )
            if self.frange is not None:
                weight = weight[self.frange[0] : self.frange[1]].contiguous()
            weight = self.pad_out(weight)
            self.inner = LinearFP16(
                self.in_features,
                self.out_features,
                weight.T.contiguous(),
                None,
                self.full_in_features,
                self.full_out_features,
                self.first_in_feature,
                self.first_out_feature,
                out_dtype = self.out_dtype,
                key = self.key
            )
            self.quant_type = "fp16"
            return True

        elif self.fkey and self.config.stc.has_tensor_group(self.fkey, ["weight"]) and self.frange is not None:

            weight = self.config.stc.get_tensor(
                self.fkey + ".weight",
                self.device,
                no_defer = True
            )
            weight = weight[self.frange[0] : self.frange[1]].contiguous()
            weight = self.pad_out(weight)
            bias = self.config.stc.get_tensor(
                self.fkey + ".bias",
                self.device,
                optional = True,
                no_defer = True
            )
            if bias is not None:
                bias = bias[self.frange[0] : self.frange[1]].contiguous()
                bias = self.pad_out(bias)
            self.inner = LinearFP16(
                self.in_features,
                self.out_features,
                weight.T.contiguous(),
                bias,
                self.full_in_features,
                self.full_out_features,
                self.first_in_feature,
                self.first_out_feature,
                out_dtype = self.out_dtype,
                key = self.key
            )
            self.quant_type = "fp16"
            return True

        return False


    def is_exl3_storage(self, key: str):
        return self.config.stc.has_tensor_group(
            key,
            [["sv", "svh"], ["su", "suh"], "trellis"]
        )

    def load_exl3(self, key: str) -> bool:
        if not self.is_exl3_storage(key):
            return False
        self.used_alt_key = key == self.alt_key
        scale = self.config.stc.get_tensor(key + ".scale", self.device, optional = True)
        su = self.config.stc.get_tensor(key + ".su", self.device, optional = True, no_defer = True)
        suh = self.config.stc.get_tensor(key + ".suh", self.device, optional = True)
        sv = self.config.stc.get_tensor(key + ".sv", self.device, optional = True, no_defer = True)
        svh = self.config.stc.get_tensor(key + ".svh", self.device, optional = True)
        trellis = self.config.stc.get_tensor(key + ".trellis", self.device)
        # TODO: We technically don't need to load these unless we need to save the tensors later
        mcg = self.config.stc.get_tensor(key + ".mcg", "cpu", optional = True)
        mul1 = self.config.stc.get_tensor(key + ".mul1", "cpu", optional = True)
        bias = self.config.stc.get_tensor(key + ".bias", self.device, optional = True)
        self.inner = LinearEXL3(
            self.config,
            self.in_features,
            self.out_features,
            scale,
            su,
            sv,
            suh,
            svh,
            trellis,
            mcg,
            mul1,
            bias,
            self.out_dtype,
            key = self.key
        )
        self.quant_type = "exl3"
        return True


    @override
    def can_defer_load(self):
        if self.is_sliced: return False
        return super().can_defer_load()


    @override
    def load(self, device: torch.device, **kwargs):
        self.device = device
        keys = [self.key]
        if self.alt_key:
            keys += [self.alt_key]
        if any(self.load_exl3(k) for k in keys): return
        if any(self.load_fp16(k) for k in keys): return
        raise ValueError(f"No tensors found for {self.key} matching supported quantization format.")


    @override
    def unload(self):
        if self.inner is not None:
            self.inner.unload()
        self.device = None
        self.inner = None


    @override
    def get_tensors(self):
        if self.device:
            return self.inner.get_tensors(self.key)
        else:
            return {}


    def convert_exl3(
        self,
        H_data: dict,
        quant_args: dict,
        progress_str: str | None = None,
        return_weight_q: bool = False,
        verbose: bool = False,
        save_reg: str = None,
        override_swap_device: torch.device | None = None
    ):
        assert isinstance(self.inner, LinearFP16), \
            "Inner layer is already quant type"

        # Destroy original layer here to save VRAM, we only need weights
        swap_to_device = self.inner.swap_device  # in case weights are swapped to CPU
        if swap_to_device is not None and override_swap_device is not None:
            swap_to_device = override_swap_device

        orig_weight = self.inner.get_weight_tensor().float()
        orig_bias = self.inner.get_bias_tensor()
        self.inner = None

        weight_q, proxy_err, out_tensors = quantize_exl3(
            orig_weight,
            H_data,
            quant_args,
            return_weight_q,
            progress_str,
            verbose,
            swap_to_device,
            save_reg = save_reg
        )

        self.inner = LinearEXL3(
            self.config,
            self.in_features,
            self.out_features,
            out_tensors.get("scale"),
            out_tensors.get("su"),
            out_tensors.get("sv"),
            out_tensors.get("suh"),
            out_tensors.get("svh"),
            out_tensors.get("trellis"),
            out_tensors.get("mcg"),
            out_tensors.get("mul1"),
            orig_bias,
            self.out_dtype,
            key = self.key
        )

        if return_weight_q:
            return proxy_err, weight_q
        else:
            return proxy_err


    def capture_H(self, x: torch.Tensor, params: dict):
        if self.qmap not in params["capture"]:
            params["capture"][self.qmap] = {
                "H": torch.zeros(self.in_features, self.in_features, dtype = torch.float32, device = self.device),
                "first_key": self.key,
                "count": 0,
                "finalized": False,
                "num_total": 0,
                "inf_nan": torch.zeros(2, dtype = torch.long, device = self.device),
            }

        params["capture"][self.qmap]["num_total"] += x.numel()
        ext.count_inf_nan(x, params["capture"][self.qmap]["inf_nan"])

        if params["capture"][self.qmap]["first_key"] == self.key:
            rows = np.prod(x.shape[:-1])
            dim = x.shape[-1]
            x = x.view((rows, dim)).to(torch.float, copy = True)  # TODO: Why copy here?

            params["capture"][self.qmap]["H"].addmm_(x.T, x)
            params["capture"][self.qmap]["count"] += rows


    @override
    def weights_numel(self):
        return self.in_features * self.out_features


    @override
    def forward(
        self,
        x: torch.Tensor,
        params: dict,
        out_dtype: torch.dtype | None = None,
    ) -> torch.Tensor:

        if "capture" in params and self.qmap:
            self.capture_H(x, params)

        x = self.inner.forward(x, params, out_dtype)
        if self.softcap != 0.0:
            ext.softcap(x, x, self.softcap)
        if self.post_scale != 1.0:
            x *= self.post_scale
        return x


    def allocate_q(self, quant_args: dict, surplus_bits: int):
        return allocate_linear(
            quant_args[self.qbits_key],
            surplus_bits,
            self
        )


    def quant_format_id(self):
        # alt_key is only used when loading unquantized model
        if self.is_exl3_storage(self.key):
            return "exl3"
        else:
            return None


    @cached_property
    def _storage_size(self):
        # alt_key is only used when loading unquantized model
        if self.is_exl3_storage(self.key):
            return sum(self.config.stc.get_tensor_sizes(prefix = self.key))
        else:
            return 2 * self.in_features * self.out_features
    def storage_size(self):
        return self._storage_size


    def recons_size(self):
        return 2 * self.in_features * self.out_features


    def make_tp_allocation(self, options: dict) -> list[TPAllocation]:
        storage = 0
        storage += self.storage_size()
        overhead_d = self.out_features * (self.out_dtype or torch.half).itemsize
        overhead_s = self.out_features * (self.out_dtype or torch.half).itemsize
        recons = self.recons_size()
        tpa = TPAllocation(
            key = self.key,
            channel_width = 128,
            channel_unit = "channels",
            storage_per_device = 0,
            storage_to_split = storage,
            overhead_per_device = overhead_d,
            overhead_to_split = overhead_s,
            recons_temp = recons,
            channels_to_split = self.out_features // 128,
            limit_key = "linear"
        )
        return [tpa]


    def tp_export(self, plan, producer):
        assert self.device is not None and self.inner is not None, "Cannot export module for TP before loading."
        return {
            "cls": Linear,
            "kwargs": {
                "key": self.key,
                "in_features": self.in_features,
                "out_features": self.out_features,
                "out_dtype": self.out_dtype,
                "alt_key": self.alt_key,
                "caps": self.caps,
                "softcap": self.softcap,
                "full_in_features": self.full_in_features,
                "full_out_features": self.full_out_features,
                "first_in_feature": self.first_in_feature,
                "first_out_feature": self.first_out_feature,
                "post_scale": self.post_scale,
            },
            "inner": self.inner.tp_export(plan, producer),
            "device": self.device
        }


    @staticmethod
    def tp_import_split(local_context, exported, plan, split):
        device = local_context["device"]
        module = Linear(
            config = None,
            **exported["kwargs"],
        )
        module.device = device
        module.inner = exported["inner"]["cls"].tp_import_split(local_context, exported["inner"], plan, split)
        module.quant_type = module.inner.quant_type
        module.out_features = module.inner.out_features
        return module

    @staticmethod
    def tp_import(local_context, exported, plan):
        key = exported["kwargs"]["key"]
        first, last, unit = plan[key] if key in plan else (None, None, "channels")
        assert unit == "channels"
        split = (True, first, last) if first is not None else None
        return Linear.tp_import_split(local_context, exported, plan, split)



from __future__ import annotations
from typing import override
import torch
from torch import nn
from ..model.config import Config
from ..util.tensor import to2
from . import Module
from ..tokenizer.mm_embedding import FIRST_MM_EMBEDDING_INDEX
from ..model.model_tp_alloc import TPAllocation

class Embedding(Module):

    def __init__(
        self,
        config: Config | None,
        key: str,
        vocab_size: int,
        hidden_size: int,
        out_dtype: torch.dtype | None = torch.float,
        qmap: str | None = None,
        normalize: bool = False,
        multiplier: float = 1.0
    ):
        super().__init__(config, key, None)
        assert qmap is None, "No quant scheme for Embedding"

        self.key = key
        self.embedding = None
        self.vocab_size = vocab_size
        self.hidden_size = hidden_size
        self.out_dtype = out_dtype
        self._numel = vocab_size * hidden_size
        self.normalize = normalize
        self.multiplier = multiplier

        self.caps.update({
            "prefer_cpu": True,
        })

    @override
    def optimizer_targets(self):
        return []

    @override
    def load(self, device: torch.device, **kwargs):
        self.device = device
        weight = self.config.stc.get_tensor(self.key + ".weight", self.device, float2half = True)
        self._numel = weight.numel()
        self.embedding = nn.Embedding(
            self.vocab_size,
            self.hidden_size,
            device = "meta"
        )
        self.embedding.weight = nn.Parameter(weight)

    @override
    def unload(self):
        self.device = None
        self.embedding = None

    @override
    def get_tensors(self):
       return {
            f"{self.key}.weight": self.embedding.weight.data.contiguous()
        }

    @override
    def weights_numel(self):
        return self._numel
        
    @override
    def forward(
        self,
        x: torch.Tensor,
        params: dict,
        out_dtype: torch.dtype | None = None
    ) -> torch.Tensor:

        indexed_emb = params.get("indexed_embeddings")
        input_ids = x
        out_dtype = out_dtype or self.out_dtype or x.dtype

        # Indexed embedding masks
        if indexed_emb:
            standard_mask = input_ids < FIRST_MM_EMBEDDING_INDEX
            indexed_masks = [
                (input_ids >= e.first_index) & (input_ids < (e.first_index + e.mm_length))
                for e in indexed_emb
            ]
            indexed_act = [im.any() for im in indexed_masks]
            use_indexed_emb = any(indexed_act)

        # Mixed embeddings when needed
        if indexed_emb and use_indexed_emb:
            bsz, seq_len = input_ids.shape
            combined_emb = torch.empty((bsz, seq_len, self.hidden_size), device = self.device, dtype = out_dtype)

            # Prepare deepstack embedding tensors
            if any(ie.deepstack_embeddings is not None for ie in indexed_emb) and indexed_act:
                assert all(ie.deepstack_embeddings is not None for ie in indexed_emb)
                num_layers = len(indexed_emb[0].deepstack_embeddings)
                assert all(num_layers == len(ie.deepstack_embeddings) is not None for ie in indexed_emb)
                deepstack_emb = [torch.zeros_like(combined_emb) for _ in range(num_layers)]
            else:
                deepstack_emb = None

            # Insert standard embeddings
            if standard_mask.any():
                for i in range(bsz):
                    standard_ids_row = input_ids[i][standard_mask[i]]
                    standard_emb_row = self.embedding(standard_ids_row)
                    combined_emb[i][standard_mask[i]] = standard_emb_row.to(out_dtype)

            # Only normalize standard embeddings
            if self.normalize:
                combined_emb *= combined_emb.shape[-1] ** 0.5

            # Insert indexed embeddings
            for im, ie, act in zip(indexed_masks, indexed_emb, indexed_act):
                if not act:
                    continue
                for i in range(bsz):
                    indexed_ids_row = input_ids[i][im[i]] - ie.first_index
                    combined_emb[i][im[i]] = ie.embeddings[indexed_ids_row].to(out_dtype)

                    # Prepare deepstack embeddings
                    if ie.deepstack_embeddings is not None:
                        for layer, de in enumerate(ie.deepstack_embeddings):
                            deepstack_emb[layer][i][im[i]] = de[indexed_ids_row].to(out_dtype)

            # Save deepstack embeddings to params
            if deepstack_emb is not None:
                params["deepstack_emb"] = deepstack_emb

            return combined_emb if self.multiplier == 1.0 else combined_emb * self.multiplier

        # No indexed embeddings, or none in current batch
        else:
            x = self.embedding.forward(x)
            x = to2(x, out_dtype, self.out_dtype)
            if self.normalize:
                x *= x.shape[-1] ** 0.5
            return x if self.multiplier == 1.0 else x * self.multiplier

    def make_tp_allocation(self, options: dict) -> list[TPAllocation]:
        return []

    def tp_export(self, plan, producer):
        assert self.device is not None, "Cannot export module for TP before loading."
        return {
            "cls": Embedding,
            "kwargs": {
                "key": self.key,
                "vocab_size": self.vocab_size,
                "hidden_size": self.hidden_size,
                "out_dtype": self.out_dtype,
                "normalize": self.normalize,
            },
            "embedding.weight": producer.send(self.embedding.weight),
            "device": self.device
        }

    @staticmethod
    def tp_import(local_context, exported, plan):
        consumer = local_context["consumer"]
        module = Embedding(
            config = None,
            **exported["kwargs"],
        )
        module.device = exported["device"]
        module.embedding = nn.Embedding(
            module.vocab_size,
            module.hidden_size,
            device = "meta"
        )
        emb = consumer.recv(exported["embedding.weight"], cuda = False)
        module.embedding.weight = nn.Parameter(emb)
        return module